"""
This module contains some fast utility functions that are useful in the same contexts as camb. They are entirely
independent of the main camb code.

"""

from ctypes import c_int, c_double, c_bool, POINTER
from .baseconfig import camblib, numpy_1d, numpy_2d, numpy_3d
import numpy as np

_chi2 = camblib.__mathutils_MOD_getchisquared
_chi2.argtypes = [numpy_2d, numpy_1d, POINTER(c_int)]
_chi2.restype = c_double


def chi_squared(covinv, x):
    """
    Utility function to efficiently calculate x^T covinv x

    :param covinv: symmetric inverse covariance matrix
    :param x: vector
    :return: covinv.dot(x).dot(x), but parallelized and using symmetry
    """
    if len(x) != covinv.shape[0] or covinv.shape[0] != covinv.shape[1]:
        raise ValueError('Wrong shape in chi_squared')
    return _chi2(covinv, x, c_int(len(x)))


int_arg = POINTER(c_int)
_3j = camblib.__mathutils_MOD_getthreejs
_3j.argtypes = [numpy_1d, int_arg, int_arg, int_arg, int_arg]


def threej(l2, l3, m2, m3):
    """
    Convenience wrapper around standard 3j function, returning array for all allowed l1 values

    :param l2: L_2
    :param l3: L_3
    :param m2: M_2
    :param m3: M_3
    :return: array of 3j from  max(abs(l2-l3),abs(m2+m3)) .. l2+l3
    """
    l1min = max(np.abs(l2 - l3), np.abs(m2 + m3))
    result = np.zeros(int(l3 + l2 - l1min + 1))
    l2in, l3in, m2in, m3in = c_int(l2), c_int(l3), c_int(m2), c_int(m3)
    _3j(result, l2in, l3in, m2in, m3in)
    return result


def threej_pt(l1, l2, l3, m1, m2, m3):
    """
    Convenience testing function to get 3j for specific arguments.
    Normally use threej to get an array at once for same cost.

    :param l1: L_1
    :param l2: L_2
    :param l3: L_3
    :param m1: M_1
    :param m2: M_2
    :param m3: M_3
    :return: Wigner 3j (integer zero if outside triangle constraints)
    """
    if m1 + m2 + m3:
        return 0
    l1min = max(np.abs(l2 - l3), np.abs(m1))
    if l1 < l1min or l1 > l2 + l3:
        return 0
    wigner = threej(l2, l3, m2, m3)
    return wigner[l1 - l1min]


# Utils_3j_integrate(W,lmax_w, n, dopol, M, lmax)
_coupling_3j = camblib.__mathutils_MOD_integrate_3j
_coupling_3j.argtypes = [numpy_2d, POINTER(c_int), POINTER(c_int), POINTER(c_bool), numpy_3d, POINTER(c_int)]


def threej_coupling(W, lmax, pol=False):
    r"""
    Calculate symmetric coupling matrix :math`\Xi` for given weights :math:`W_{\ell}`,
    where :math:`\langle\tilde{C}_\ell\rangle = \Xi_{\ell \ell'} (2\ell'+1) C_\ell`.
    The weights are related to the power spectrum of the mask P
    by :math:`W_\ell = (2 \ell + 1) P_\ell / 4 \pi`.
    See e.g. Eq D16 of `arxiv:0801.0554 <http://arxiv.org/abs/0801.0554>`_.

    If pol is False and W is an array of weights, produces array of temperature couplings, otherwise for pol is True
    produces set of TT, TE, EE, EB couplings (and weights must have one spectrum - for same masks - or three).

    Use :func:`scalar_coupling_matrix` or :func:`pcl_coupling_matrix` to get the coupling matrix directly from the
    mask power spectrum.

    :param W: 1d array of Weights for each L, or list of arrays of weights (zero based)
    :param lmax: lmax for the output matrix (assumed symmetric, though not in principle)
    :param pol: if pol, produce TT, TE, EE, EB couplings for three input mask weights (or one if assuming same mask)
    :return: symmetric coupling matrix or array of matrices
    """
    if not isinstance(W, (list, tuple)):
        W = [W]
    if pol:
        n = 4
        if len(W) == 1:
            W = W * 3
        assert len(W) == 3
    else:
        n = len(W)
    M = np.zeros((n, lmax + 1, lmax + 1))
    nW = len(W)
    lmax_w = min(2 * lmax, len(W[0]) - 1)
    for m in W[1:]:
        assert lmax_w == min(2 * lmax, len(m) - 1)
    Wmat = np.empty((nW, lmax_w + 1))
    for i, m in enumerate(W):
        Wmat[i, :] = m[:lmax_w + 1]
    _coupling_3j(Wmat, c_int(lmax_w), c_int(nW), c_bool(pol), M, c_int(lmax))
    if n == 1:
        return M[0, :, :]
    else:
        return [M[i, :, :] for i in range(n)]


def scalar_coupling_matrix(P, lmax):
    """
    Get scalar Pseudo-Cl coupling matrix from power spectrum of mask, or array of power masks.
    Uses multiple threads. See Eq A31 of `astro-ph/0105302 <https://arxiv.org/abs/astro-ph/0105302>`_

    :param P: power spectrum of mask, or list of mask power spectra
    :param lmax: lmax for the matrix (assumed square)
    :return: coupling matrix (square but not symmetric), or list of couplings for different masks
    """

    if not isinstance(P, (list, tuple)):
        P = [P]
    elif any(x.size != P[0].size for x in P[1:]):
        raise ValueError('Mask power spectra must have same lmax')

    lmax_power = min(P[0].size - 1, 2 * lmax)
    if lmax_power < 2 * lmax:
        print('Warning: power spectrum lmax is less than 2*lmax')

    fac = (2 * np.arange(lmax_power + 1) + 1) / 4 / np.pi
    M = threej_coupling([fac * power for power in P], lmax)
    factor = 2 * np.arange(lmax + 1) + 1
    if len(P) == 1:
        return M * factor
    else:
        return [m * factor for m in M]


def pcl_coupling_matrix(P, lmax, pol=False):
    """
    Get Pseudo-Cl coupling matrix from power spectrum of mask.
    Uses multiple threads. See Eq A31 of `astro-ph/0105302 <https://arxiv.org/abs/astro-ph/0105302>`_

    :param P: power spectrum of mask
    :param lmax: lmax for the matrix
    :param pol: whether to calculate TE, EE, BB couplings
    :return: coupling matrix (square but not symmetric), or list of TT, TE, EE, BB if pol
    """

    lmax_power = min(P.size - 1, 2 * lmax)
    if lmax_power < 2 * lmax:
        print('Warning: power spectrum lmax is less than 2*lmax')

    W = (2 * np.arange(lmax_power + 1) + 1) * P / (4 * np.pi)
    M = threej_coupling(W, lmax, pol=pol)

    factor = 2 * np.arange(lmax + 1) + 1
    if pol:
        return [mat * factor for mat in M]
    else:
        return M * factor


_gauss_legendre = camblib.__mathutils_MOD_gauss_legendre
_gauss_legendre.argtypes = [numpy_1d, numpy_1d, int_arg]


def gauss_legendre(xvals, weights, npoints):
    _gauss_legendre(xvals, weights, c_int(npoints))



import os
import sys
import unittest
import numpy as np

try:
    import camb
except ImportError:
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))
    import camb


class HMcodeTest(unittest.TestCase):

    def testHMcode(self):

        # Parameters
        # TODO: Not clear that the neutrino setup here agrees exactly with my HMx setup
        # TODO: Larger errors than I would like with high-massive-nu and high-baryon cosmologies
        # TODO: Normalisation via As not agreeing exactly with sigma_8 values
        kmax_calc = 2e2  # Maximum wavenumber for the CAMB calculation [h/Mpc]
        pi = np.pi  # Lovely pi
        twopi = 2. * pi  # Twice as lovely pi
        neutrino_number = 94.14  # Neutrino closure mass [eV] TODO: VERY LAZY (should it be 93.14, 94.07eV?)
        eps_k = 1e-6  # Fractional error tolerated in k
        eps_a = 1e-6  # Fractional error tolerated in a
        eps_Pk = 5e-3  # Fractional error tolerated in Pk # TODO: This should agree to 1e-3 as in HMx
        kmin_test = 1e-2  # Minimum wavenumber for test [h/Mpc]
        kmax_test = 1e1  # Maximum wavenumber for test [h/Mpc]
        amin_test = 0.333  # Minimum scale factor for test
        amax_test = 1.000  # Maximum scale factor for test
        Y_He = 0.24  # Helium fraction
        T_CMB = 2.725  # CMB temperature
        nnu = 3  # Number of massive neutrinos
        neff = 3.046  # Effective number of radiative neutrinos
        verbose = False  # Verbose tests
        norm_sig8 = False  # Normalisation by sigma_8 or As?
        a_in = np.array([0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])  # Scale factors for tests

        # Dictionary for HMcode versions
        HMcode_version_ihm = {
            'mead': 51,
            'mead2016': 51,
            'mead2015': 66,
            'mead2020': 123,
            'mead2020_feedback': 124
        }

        # Read in and sort Mead benchmark data
        def read_Mead_benchmark(infile):

            # Read data and split into k, a and Pk
            data = np.loadtxt(infile)
            k = data[:, 0]
            a = a_in
            D2 = data[:, 1:]
            Pk = (D2.T / (4. * pi * (k / twopi) ** 3)).T  # Convert Delta^2(k) -> P(k)

            # Return results
            return k, a, Pk

        def HMcode_test_cosmologies(icos):

            # Common test parameters
            Om_m = 0.30
            Om_b = 0.05
            h = 0.7
            ns = 0.96
            sig8 = 0.8
            As = 1.97269e-9
            w0 = -1.
            wa = 0.
            mnu = 0.
            logT = 7.8

            # Uncommon test parameters
            if icos == 56:
                Om_m = 0.3158
                Om_b = 0.04939
                h = 0.6732
                ns = 0.96605
                sig8 = 0.8120
                As = 2.08924e-9
                w0 = -1.
                wa = 0.
                mnu = 0.06
            elif icos == 241:
                w0 = -0.7
                As = 2.46981e-9
            elif icos == 242:
                w0 = -1.3
                As = 1.75070e-9
            elif icos == 243:
                mnu = 0.3
                As = 2.35868e-9
            elif icos == 244:
                mnu = 0.9
                As = 3.43507e-9
            elif icos == 245:
                ns = 0.7
                As = 2.38515e-9
            elif icos == 246:
                ns = 1.3
                As = 1.47601e-9
            elif icos == 247:
                Om_b = 0.01
                As = 1.20028e-9
            elif icos == 248:
                Om_b = 0.1
                As = 3.88822e-9
            elif icos == 249:
                wa = 0.9
                As = 3.35538e-9
            elif icos == 250:
                logT = 7.6
            elif icos == 251:
                logT = 8.0

            # Return cosmological parameters
            return Om_m, Om_b, h, ns, sig8, As, w0, wa, mnu, logT

        # Set up a CAMB parameter set for a Mead cosmology
        def setup_HMcode_test(a, icos):

            # Get the cosmological parameters for the cosmology
            Om_m, Om_b, h, ns, sig8, As, w0, wa, mnu, logT = HMcode_test_cosmologies(icos)

            # Redshifts
            z = -1. + 1. / a

            # Derive and set CAMB parameters
            H0 = 100. * h
            wnu = mnu / neutrino_number
            Om_nu = wnu / h ** 2
            Om_c = Om_m - Om_b - Om_nu
            Om_k = 0.
            wb = Om_b * h ** 2
            wc = Om_c * h ** 2

            # Set parameters using the traditional Fortran language
            # TODO: Check carefully against my ini files
            pars = camb.CAMBparams(WantCls=False, H0=H0, ombh2=wb, omch2=wc, omnuh2=wnu, omk=Om_k, YHe=Y_He, TCMB=T_CMB,
                                   num_nu_massive=nnu,
                                   num_nu_massless=neff - nnu,
                                   nu_mass_numbers=[nnu],
                                   nu_mass_degeneracies=[neff / nnu],
                                   nu_mass_fractions=[1.],
                                   share_delta_neff=True,
                                   )
            pars.set_dark_energy(w=w0, wa=wa, dark_energy_model='ppf')
            pars.InitPower.set_params(As=As, ns=ns)

            # Set sigma_8 normalisation (wasteful)
            if norm_sig8:
                pars.set_matter_power(redshifts=[0.], kmax=kmax_calc)
                pars.NonLinear = camb.model.NonLinear_none
                results = camb.get_results(pars)
                sig8_init = results.get_sigma8()
                As = As * (sig8 / sig8_init) ** 2
                pars.InitPower.set_params(As=As, ns=ns)  # Reset As

            # Main calculation step
            pars.set_matter_power(redshifts=z, kmax=kmax_calc)
            pars.NonLinear = camb.model.NonLinear_both
            results = camb.get_results(pars)
            if verbose:
                print(pars)

            # Return results and AGN temperature (ugly, but it needs to come out)
            return results, logT

        # Get the HMcode power from CAMB
        def get_HMcode_power_from_CAMB(results, k_in, logT, HMcode_version):

            # k and z ranges for results
            kmin = k_in[0]
            kmax = k_in[-1]
            nk = len(k_in)

            # Get non-linear spectra
            results.Params.NonLinearModel.set_params(halofit_version=HMcode_version, HMCode_logT_AGN=logT)
            k, z, Pk = results.get_matter_power_spectrum(minkh=kmin, maxkh=kmax, npoints=nk)
            Pk = Pk.T[:, ::-1]
            z = np.array(z)[::-1]
            a = 1. / (1. + z)
            if verbose:
                print('sigma_8:', results.get_sigma8()[-1])

            # Return non-linear power
            return k, a, Pk

        # Input file name
        def HMcode_benchmark_file(icos, ihm):
            return 'HMcode_test_outputs/HMcode_cos%d_hm%d.txt' % (icos, ihm)

        # Whitespace
        if verbose:
            print('')

        # Loop over cosmologies
        for icos in [26, 56, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251]:

            # Init CAMB
            results, logT = setup_HMcode_test(a_in, icos)

            # Loop over HMcode versions
            for HMcode_version in ['mead2015', 'mead2016', 'mead2020', 'mead2020_feedback']:

                # Read benchmark data
                ihm = HMcode_version_ihm[HMcode_version]
                infile = HMcode_benchmark_file(icos, ihm)
                if verbose:
                    print('Infile:', infile)
                k_in, a_in, Pk_in = read_Mead_benchmark(infile)

                # Get power from CAMB
                k_nl, a_nl, Pk_nl = get_HMcode_power_from_CAMB(results, k_in, logT, HMcode_version)

                # Compare benchmark to calculation
                for ik in range(len(k_in)):
                    self.assertAlmostEqual(k_nl[ik] / k_in[ik], 1., delta=eps_k)
                for ia in range(len(a_in)):
                    self.assertAlmostEqual(a_nl[ia] / a_in[ia], 1., delta=eps_a)
                for ia in range(len(a_in)):
                    for ik in range(len(k_in)):
                        if kmin_test <= k_in[ik] <= kmax_test and amin_test <= a_in[ia] <= amax_test:
                            self.assertAlmostEqual(Pk_nl[ik, ia] / Pk_in[ik, ia], 1., delta=eps_Pk)



import os
from .baseconfig import import_property, CAMBError
from ctypes import c_char, c_int, c_bool, c_double

lensing_method_curv_corr = 1
lensing_method_flat_corr = 2
lensing_method_harmonic = 3


class _config:
    # print feedback if > 0 (note in Jupyter notebook this will appear in the terminal, not the notebook)
    FeedbackLevel = import_property(c_int, "config", "FeedbackLevel")

    # print additional timing and progress (when FeedbackLevel>0)
    DebugMsgs = import_property(c_bool, "config", "DebugMsgs")

    global_error_flag = import_property(c_int, "config", "global_error_flag")

    ThreadNum = import_property(c_int, "config", "threadnum")

    DoTensorNeutrinos = import_property(c_bool, "gaugeinterface", "dotensorneutrinos")

    DebugParam = import_property(c_double, "config", "debugparam")

    lensing_method = import_property(c_int, "lensing", "lensing_method")

    lensing_sanity_check_amplitude = import_property(c_double, "lensing", "lensing_sanity_check_amplitude")
    # lensing_sanity_check_amplitude.value = 1e-7 by default, will error if  (2*L+1)L(L+1)/4pi C_phi_phi > lensing_
    # sanity_check_amplitude at L=10
    # increase to large number to prevent sanity check (but lensing requires realistic amplitude as non-linear)

    lensing_includes_tensors = import_property(c_bool, "lensing", "lensing_includes_tensors")

    transfer_power_var = import_property(c_int, "transfer", "transfer_power_var")

    _global_error_message = import_property(c_char * 1024, "config", "global_error_message")

    def global_error_message(self):
        return bytearray(self._global_error_message).decode('ascii').strip()

    def check_global_error(self, reference=''):
        code = self.global_error_flag
        if code:
            err = config.global_error_message()
            self.global_error_flag = 0
            if reference:
                reference = 'Error in Fortran called from %s:\n' % reference
            else:
                reference = ''
            if err:
                raise CAMBError(reference + '%s' % err)
            else:
                raise CAMBError(reference + 'Error code: %s' % code)

    def __repr__(self):
        s = ''
        for x in dir(self):
            if x[0] != '_':
                value = getattr(self, x)
                if not callable(value):
                    s += '%s = %s\n' % (x, value)
        return s


config = _config()

if os.environ.get('BINDER_LAUNCH_HOST'):
    config.ThreadNum = 1  # binder is very slow with more than 1 CPU, force 1 by default



import sys
import subprocess
import os
import shutil
from typing import Any
from setuptools import setup, Command, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.build_py import build_py
from setuptools.command.develop import develop
from setuptools.command.install import install
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

file_dir = os.path.abspath(os.path.dirname(__file__))
os.chdir(file_dir)

sys.path.insert(0, os.path.join(file_dir, 'camb'))
_compile: Any = __import__('_compilers')

if _compile.is_windows:
    DLLNAME = 'cambdll.dll'
else:
    DLLNAME = 'camblib.so'


def get_forutils():
    fpath = os.getenv('FORUTILSPATH')

    def git_install_forutils():
        try:
            print('forutils not found, attempting to install using git...')
            os.chdir('..')
            fbranch = os.getenv('FORUTILSBRANCH', '1.0.3' if os.environ.get('CONDA_BUILD') else 'master')
            try:
                if subprocess.call("git clone --branch %s --depth=1 https://github.com/cmbant/forutils" % fbranch,
                                   shell=True) == 0:
                    return os.path.join('..', 'forutils')
            finally:
                os.chdir('fortran')
        except Exception:
            print('Failed to install using git')

    if not fpath:
        dirs = ['..', '..' + os.sep + '..']
        for _dir in dirs:
            path = os.path.join(_dir, 'forutils')
            if os.path.isdir(path):
                fpath = path
                main_dir = _dir
                break
        if not fpath:
            fpath = git_install_forutils()
        elif not os.path.exists(os.path.join(fpath, 'Makefile')):
            if os.path.isdir(os.path.join('..', '.git')):
                # submodule may not be pulled
                try:
                    try:
                        os.chdir('..')
                        print('forutils directory found but no Makefile. Attempting to clone submodule...')
                        if subprocess.call("git submodule update --init --recursive", shell=True, cwd=main_dir) != 0:
                            raise Exception()
                    finally:
                        os.chdir('fortran')
                except Exception:
                    fpath = None
                    print('Failed to install forutils using git')
            else:
                os.rmdir(fpath)
                if not os.path.exists(fpath):
                    fpath = git_install_forutils()
                else:
                    fpath = None

    if not fpath:
        raise Exception('Install forutils from https://github.com/cmbant/forutils, '
                        'pull the forutils submodule, or set FORUTILSPATH variable.\n'
                        'If you are cloning with git, use "git clone --recursive"')
    return fpath


def clean_dir(path, rmdir=False):
    if os.path.isdir(path):
        for f in os.listdir(path):
            os.remove(os.path.join(path, f))
        if rmdir:
            os.rmdir(path)


def make_library(cluster=False):
    os.chdir(os.path.join(file_dir, 'fortran'))
    pycamb_path = '..'
    lib_file = os.path.join(pycamb_path, 'camb', DLLNAME)
    if _compile.is_windows or not _compile.check_ifort():
        ok, gfortran_version = _compile.check_gfortran(msg=not _compile.is_windows)
        if ok and '8.2.0' in gfortran_version:
            print('WARNING: gfortran 8.2.0 may be buggy and give unreliable results or crashes, upgrade gfortran.')
    if _compile.is_windows:
        COMPILER = "gfortran"
        FFLAGS = "-shared -static -cpp -fopenmp -O3 -fmax-errors=4"
        # FFLAGS = "-shared -static -cpp -fopenmp -g -fbacktrace -ffpe-trap=invalid,overflow,zero " \
        #         "-fbounds-check -fmax-errors=4"
        if _compile.is_32_bit:
            FFLAGS = "-m32 " + FFLAGS
        if not ok:
            print('WARNING: gfortran %s or higher not in path (if you just installed '
                  'you may need to log off and on again).' % _compile.gfortran_min)
            print('        You can get a Windows gfortran build from https://sourceforge.net/projects/mingw-w64/files/')
            print('        - go to Files, and download MinGW-W64 Online Installer.')
            print('        Alternatively newer versions at https://github.com/niXman/mingw-builds-binaries')
            if _compile.is_32_bit:
                raise IOError('No 32bit Windows DLL provided, you need to build or use 64 bit python')
            else:
                print('Using pre-compiled binary instead - any local changes will be ignored...')
        else:
            fpath = get_forutils()
            makefile = _compile.makefile_dict('Makefile_main')
            SOURCES = makefile['SOURCEFILES'].split()
            FORUTILS = [os.path.join(fpath, f.replace('.f90', '')) for f in
                        _compile.makefile_dict(os.path.join(fpath, 'Makefile'))['SRCS'].replace('MatrixUtils.f90',
                                                                                                '').split()]
            tmpdir = 'WinDLL'
            if not os.path.isdir(tmpdir):
                os.mkdir(tmpdir)
            ofiles = []
            new_compiler = True
            ver_file = os.path.join(tmpdir, 'compiler.ver')
            if os.path.exists(ver_file):
                with open(ver_file, 'r') as f:
                    new_compiler = gfortran_version != f.readline().strip()
            if new_compiler:
                clean_dir(tmpdir)
                with open(ver_file, 'w') as f:
                    f.write(gfortran_version)

            need_compile = not os.path.exists(lib_file)
            if not need_compile:
                dll_time = os.path.getmtime(lib_file)
            for source in FORUTILS + SOURCES:
                # manual Make using dependency files if available
                outroot = os.path.join(tmpdir, os.path.split(source)[1])
                fout = outroot + '.o'
                ofiles += [fout]
                modified = new_compiler or not os.path.exists(fout)
                if not modified:
                    o_time = os.path.getmtime(fout)
                    modified = o_time < os.path.getmtime(source + '.f90') or not os.path.exists(outroot + '.d')
                    if not modified:
                        with open(outroot + '.d', 'r') as f:
                            for dependence in " ".join(f.readlines()).replace("\\\n", "").split(':')[1].strip().split():
                                if os.path.getmtime(dependence) > o_time:
                                    modified = True
                                    break

                if modified:
                    need_compile = True
                    cmd = COMPILER + ' ' + FFLAGS + ' ' + source + '.f90 -MMD -c -o %s -J%s' % (fout, tmpdir)
                    print(cmd)
                    if subprocess.call(cmd, shell=True, env=_compile.compiler_environ) != 0:
                        raise IOError('Compilation failed')
                elif not need_compile and dll_time < o_time:
                    need_compile = True

            if need_compile or not os.path.exists(lib_file):
                if os.path.exists(lib_file):
                    # raise an exception if the file in use and cannot be deleted
                    try:
                        os.remove(lib_file)
                    except OSError:
                        raise IOError('dll file in use. Stop python codes and notebook kernels that are using camb.')
                print('Compiling sources...')
                cmd = COMPILER + ' ' + FFLAGS + ' ' + " ".join(ofiles) + ' -o %s -J%s' % (lib_file, tmpdir)
                print(cmd)
                if subprocess.call(cmd, shell=True, env=_compile.compiler_environ) != 0:
                    raise IOError('Compilation failed')
            else:
                print('DLL up to date.')
    else:
        if not _compile.call_command('make -v'):
            raise IOError('Build failed - you must have "make" installed. '
                          'E.g. on ubuntu install with "sudo apt install make" (or use build-essential package).')
        get_forutils()
        print("Compiling source...")
        subprocess.call("make python PYCAMB_OUTPUT_DIR=%s/camb/ CLUSTER_SAFE=%d" %
                        (pycamb_path, int(cluster if not os.getenv("GITHUB_ACTIONS") else 1)), shell=True)
        subprocess.call("chmod 755 %s" % lib_file, shell=True)

    if not os.path.isfile(os.path.join(pycamb_path, 'camb', DLLNAME)):
        sys.exit('Compilation failed')
    tem_file = 'HighLExtrapTemplate_lenspotentialCls.dat'
    tem = os.path.join(pycamb_path, 'camb', tem_file)
    if not os.path.exists(tem) or os.path.getmtime(tem) < os.path.getmtime(tem_file):
        shutil.copy(tem_file, tem)

    os.chdir(file_dir)


class MakeLibrary(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        make_library(False)


class MakeLibraryCluster(MakeLibrary):

    def run(self):
        make_library(True)


class SharedLibrary(build_py):

    def run(self):
        make_library(False)
        build_py.run(self)


class SharedLibraryCluster(SharedLibrary):

    def run(self):
        make_library(True)
        build_py.run(self)


class DevelopLibrary(develop):

    def run(self):
        make_library(False)
        develop.run(self)


class DevelopLibraryCluster(develop):

    def run(self):
        make_library(True)
        develop.run(self)


class CleanLibrary(MakeLibrary):

    def run(self):
        if _compile.is_windows:
            clean_dir(os.path.join(file_dir, 'fortran', 'WinDLL'), rmdir=True)
        else:
            subprocess.call("make clean", shell=True, cwd=os.path.join(file_dir, 'fortran'))


class BDistWheelNonPure(_bdist_wheel):
    def finalize_options(self):
        super().finalize_options()
        self.root_is_pure = False

    def get_tag(self):
        _, _, plat = super().get_tag()
        if "osx_12" in plat:
            return _, _, plat
        return "py3", "none", plat


class InstallPlatlib(install):
    def finalize_options(self):
        super().finalize_options()
        if self.distribution.has_ext_modules():
            self.install_lib = self.install_platlib


class BuildExtCommand(build_ext):
    """Ensure built extensions are added to the correct path in the wheel."""

    def run(self):
        pass


if __name__ == "__main__":
    setup(name=os.getenv('CAMB_PACKAGE_NAME', 'camb'),
          zip_safe=False,
          cmdclass={'build_py': SharedLibrary, 'build_cluster': SharedLibraryCluster,
                    'make': MakeLibrary, 'make_cluster': MakeLibraryCluster, 'clean': CleanLibrary,
                    'develop': DevelopLibrary, 'develop_cluster': DevelopLibraryCluster,
                    'bdist_wheel': BDistWheelNonPure, 'install': InstallPlatlib,
                    "build_ext": BuildExtCommand},
          ext_modules=[Extension("camb.camblib", [])],
          packages=['camb', 'camb.tests'],
          platforms="any",
          package_data={'camb': [DLLNAME, 'HighLExtrapTemplate_lenspotentialCls.dat',
                                 'PArthENoPE_880.2_marcucci.dat', 'PArthENoPE_880.2_standard.dat',
                                 'PRIMAT_Yp_DH_Error.dat', 'PRIMAT_Yp_DH_ErrorMC_2021.dat']},
          test_suite='camb.tests'
          )



from .baseconfig import camblib, CAMBError, CAMBValueError, CAMBUnknownArgumentError, CAMB_Structure, F2003Class, \
    fortran_class, numpy_1d, numpy_2d, numpy_1d_int, fortran_array, AllocatableArrayDouble, ndpointer, np, lib_import
from ctypes import c_float, c_int, c_double, c_bool, POINTER, byref
import ctypes
from . import model, constants
from ._config import config
from .model import set_default_params, CAMBparams
import logging
from scipy.interpolate import RectBivariateSpline, interp1d

int_arg = POINTER(c_int)
d_arg = POINTER(c_double)


class _MatterTransferData(CAMB_Structure):
    # contains complex types with pointers, so just set up dummy
    _fields_ = [('num_q_trans', c_int),
                ('q_trans', POINTER(c_double)),
                ('sigma_8', POINTER(c_double)),
                ('sigma2_vdelta_8', POINTER(c_double)),
                ('TransferData', POINTER(c_float)),
                ('sigma_8_size', c_int),
                ('sigma2_vdelta_8_size', c_int),
                ('TransferData_size', c_int * 3)
                ]


class _ClTransferData(CAMB_Structure):
    _fields_ = [('NumSources', c_int),
                ('q_size', c_int),
                ('q', POINTER(c_double)),
                ('delta_size', c_int * 3),
                ('delta_p_l_k', POINTER(c_double)),
                ('l_size', c_int),
                ('L', POINTER(c_int))
                ]


def save_cmb_power_array(filename, array, labels, lmin=0):
    """
    Save a zero-based 2-d array of CL to a text file, with each line starting with L.

    :param filename: filename to save
    :param array: 2D array of power spectra
    :param labels:  header names for each column in the output
    :param lmin: L to start output in file (usually 0 or 2)
    """
    lmax = array.shape[0] - 1
    ls = np.atleast_2d(np.arange(lmin, lmax + 1)).T
    ncol = array.shape[1]
    if isinstance(labels, str):
        labels = labels.split()
    # noinspection PyTypeChecker
    np.savetxt(filename, np.hstack((ls, array[lmin:, :])), fmt=['%4u'] + ['%12.7e'] * ncol,
               header=' L ' + ' '.join(['{:13s}'.format(lab) for lab in labels]))


class MatterTransferData:
    r"""
    MatterTransferData is the base class for storing matter power transfer function data for various q values.
    In a flat universe q=k, in a closed universe q is quantized.

    To get an instance of this data, call :meth:`.results.CAMBdata.get_matter_transfer_data`.

    For a description of the different Transfer_xxx outputs (and 21cm case) see :ref:`transfer-variables`; the
    array is indexed by index+1 given by:

    - Transfer_kh = 1 (k/h)
    - Transfer_cdm = 2 (cdm)
    - Transfer_b = 3 (baryons)
    - Transfer_g = 4 (photons)
    - Transfer_r = 5 (massless neutrinos)
    - Transfer_nu = 6 (massive neutrinos)
    - Transfer_tot = 7 (total matter)
    - Transfer_nonu = 8 (total matter excluding neutrinos)
    - Transfer_tot_de = 9 (total including dark energy perturbations)
    - Transfer_Weyl = 10 (Weyl potential)
    - Transfer_Newt_vel_cdm = 11 (Newtonian CDM velocity)
    - Transfer_Newt_vel_baryon = 12 (Newtonian baryon velocity)
    - Transfer_vel_baryon_cdm = 13 (relative baryon-cdm velocity)


    :ivar nq:  number of q modes calculated
    :ivar q: array of q values calculated
    :ivar sigma_8: array of :math:`\sigma_8` values for each redshift
    :ivar sigma2_vdelta_8: array of v-delta8 correlation, so sigma2_vdelta_8/sigma_8 can define growth
    :ivar transfer_data: numpy array T[entry, q_index, z_index] storing transfer functions for each redshift and q;
                         entry+1 can be one of the Transfer_xxx variables above.
    """

    nq: int
    q: np.ndarray
    sigma_8: np.ndarray
    sigma2_vdelta_8: np.ndarray
    transfer_data: np.ndarray

    def transfer_z(self, name, z_index=0):
        """
        Get transfer function (function of q, for each q in self.q_trans) by name for given redshift index

        :param name:  parameter name
        :param z_index: which redshift
        :return: array of transfer function values for each calculated k
        """

        if name not in model.transfer_names:
            raise CAMBError('Unknown name %s; must be one of %s' % (name, model.transfer_names))
        return self.transfer_data[model.transfer_names.index(name), :, z_index]


class ClTransferData:
    r"""
    ClTransferData is the base class for storing CMB power transfer functions, as a function of q and :math:`\ell`.
    To get an instance of this data, call :meth:`.results.CAMBdata.get_cmb_transfer_data`

    :ivar NumSources:  number of sources calculated (size of p index)
    :ivar q: array of q values calculated (=k in flat universe)
    :ivar L: int array of :math:`\ell` values calculated
    :ivar delta_p_l_k: transfer functions, indexed by source, L, q
    """
    NumSources: int
    q: np.ndarray
    L: np.ndarray
    delta_p_l_k: np.ndarray

    def get_transfer(self, source=0):
        r"""
        Return :math:`C_\ell` transfer functions as a function of :math:`\ell`
        and :math:`q` (:math:`= k` in a flat universe).

        :param source: index of source: e.g. 0 for temperature, 1 for E polarization, 2 for lensing potential
        :return: array of computed L, array of computed q, transfer functions T(L,q)
        """

        return self.L, self.q, self.delta_p_l_k[source, :, :]


@fortran_class
class CAMBdata(F2003Class):
    """
    An object for storing calculational data, parameters and transfer functions.
    Results for a set of parameters (given in a :class:`~.model.CAMBparams` instance)
    are returned by the :func:`.camb.get_background`, :func:`.camb.get_transfer_functions` or :func:`.camb.get_results`
    functions. Exactly which quantities are already calculated depends on which of these functions you use and the
    input parameters.

    To quickly make a fully calculated CAMBdata instance for a set of parameters you can call :func:`.camb.get_results`.

    """
    _fortran_class_module_ = 'results'

    _fields_ = [("Params", CAMBparams),
                ("ThermoDerivedParams", c_double * model.nthermo_derived,
                 "array of derived parameters, see :meth:`get_derived_params` to get as a dictionary"),
                ("flat", c_bool, "flat universe"),
                ("closed", c_bool, "closed universe"),
                # grho gives the contribution to the expansion rate from: (g) photons,
                # (r) one flavor of relativistic neutrino (2 degrees of freedom),
                #  grho is actually 8*pi*G*rho/c^2 at a=1, with units of Mpc**(-2).
                ("grhocrit", c_double, "kappa*a^2*rho_c(0)/c^2 with units of Mpc**(-2)"),
                ("grhog", c_double, "kappa/c^2*4*sigma_B/c^3 T_CMB^4"),
                ("grhor", c_double, "7/8*(4/11)^(4/3)*grhog (per massless neutrino species)"),
                ("grhob", c_double, "baryon contribution"),
                ("grhoc", c_double, "CDM contribution"),
                ("grhov", c_double, "Dark energy contribution"),
                ("grhornomass", c_double, "grhor*number of massless neutrino species"),
                ("grhok", c_double, "curvature contribution to critical density"),
                ("taurst", c_double, "time at start of recombination"),
                ("dtaurec", c_double, "time step in recombination"),
                ("taurend", c_double, "time at end of recombination"),
                ("tau_maxvis", c_double, "time at peak visibility"),
                ("adotrad", c_double, "da/d tau in early radiation-dominated era"),
                ("omega_de", c_double, "Omega for dark energy today"),
                ("curv", c_double, "curvature K"),
                ("curvature_radius", c_double, r":math:`1/\sqrt{|K|}`"),
                ("Ksign", c_double, "Ksign = 1,0 or -1"),
                ("tau0", c_double, "conformal time today"),
                ("chi0", c_double, "comoving angular diameter distance of big bang; rofChi(tau0/curvature_radius)"),
                ("scale", c_double, "relative to flat. e.g. for scaling L sampling"),
                ("akthom", c_double, "sigma_T * (number density of protons now)"),
                ("fHe", c_double, "n_He_tot / n_H_tot"),
                ("Nnow", c_double, "number density today"),
                ("z_eq", c_double, "matter-radiation equality redshift assuming all neutrinos relativistic"),
                ("grhormass", c_double * model.max_nu),
                ("nu_masses", c_double * model.max_nu),
                ("num_transfer_redshifts", c_int,
                 "Number of calculated redshift outputs for the matter transfer (including those for CMB lensing)"),
                ("transfer_redshifts", AllocatableArrayDouble, "Calculated output redshifts"),
                ("PK_redshifts_index", c_int * model.max_transfer_redshifts, "Indices of the requested PK_redshifts"),
                ("OnlyTransfers", c_bool, "Only calculating transfer functions, not power spectra"),
                ("HasScalarTimeSources", c_bool, "calculate and save time source functions, not power spectra")]

    # Note there are many more fields in Fortran. Since F2003Class is memory-managed by Fortran, we don't need
    # need to define them all in python.
    # _methods_ refer to the imported functions in the corresponding fortran class.

    _methods_ = [('AngularDiameterDistance', [d_arg], c_double),
                 ('AngularDiameterDistanceArr', [numpy_1d, numpy_1d, int_arg]),
                 ('AngularDiameterDistance2', [d_arg, d_arg], c_double),
                 ('AngularDiameterDistance2Arr', [numpy_1d, numpy_1d, numpy_1d, int_arg]),
                 ('ComovingRadialDistance', [d_arg], c_double),
                 ('ComovingRadialDistanceArr', [numpy_1d, numpy_1d, int_arg, d_arg]),
                 ('Hofz', [d_arg], c_double),
                 ('HofzArr', [numpy_1d, numpy_1d, int_arg]),
                 ('DeltaPhysicalTimeGyr', [d_arg, d_arg, d_arg], c_double),
                 ('DeltaPhysicalTimeGyrArr', [numpy_1d, numpy_1d, numpy_1d, int_arg, d_arg]),
                 ('GetBackgroundDensities', [int_arg, numpy_1d, numpy_2d]),
                 ('DeltaTime', [d_arg, d_arg, d_arg], c_double),
                 ('DeltaTimeArr', [numpy_1d, numpy_1d, numpy_1d, int_arg, d_arg]),
                 ('TimeOfzArr', [numpy_1d, numpy_1d, int_arg, d_arg]),
                 ('sound_horizon_zArr', [numpy_1d, numpy_1d, int_arg]),
                 ('RedshiftAtTimeArr', [numpy_1d, numpy_1d, int_arg]),
                 ('CosmomcTheta', [], c_double),
                 ('DarkEnergyStressEnergy', [numpy_1d, numpy_1d, numpy_1d, int_arg]),
                 ('get_lmax_lensed', [], c_int),
                 ('get_zstar', [d_arg], c_double),
                 ('SetParams', [POINTER(CAMBparams), int_arg, int_arg, int_arg, int_arg])
                 ]

    def __init__(self):
        set_default_params(self.Params)

    def __getstate__(self):
        raise TypeError("Cannot save CAMB result objects")

    def set_params(self, params):
        """
        Set parameters from params. Note that this does not recompute anything;
        you will need to call :meth:`calc_transfers` if you change any parameters affecting the
        background cosmology or the transfer function settings.

        :param params: a :class:`~.model.CAMBparams` instance
        """
        self.Params = params

    def get_derived_params(self):
        """
        :return: dictionary of derived parameter values, indexed by name ('kd', 'age', etc..)
        """
        res = {}
        for name, value in zip(model.derived_names, self.ThermoDerivedParams):
            res[name] = value
        return res

    def get_background_outputs(self):
        """
        Get BAO values for redshifts set in Params.z_outputs

        :return: rs/DV, H, DA, F_AP for each requested redshift (as 2D array)
        """
        n = len(self.Params.z_outputs)
        if not n:
            raise CAMBError('Set z_outputs with required redshifts (and then calculate transfers/results)'
                            ' before calling get_background_outputs')
        outputs = np.empty((n, 4))
        CAMB_GetBackgroundOutputs(byref(self), outputs, byref(c_int(n)))
        return outputs

    def get_BAO(self, redshifts, params):
        """
        Get BAO parameters at given redshifts, using parameters in params

        :param redshifts: list of redshifts
        :param params: optional :class:`~.model.CAMBparams` instance to use
        :return: array of rs/DV, H, DA, F_AP for each redshift as 2D array
        """
        P = params.copy()
        P.z_outputs = redshifts
        self.calc_background(P)
        return self.get_background_outputs()

    @staticmethod
    def _check_params(params):
        if not isinstance(params, CAMBparams):
            raise CAMBValueError('Must pass a CAMBparams instance')
        if not params.ombh2:
            raise CAMBValueError('Parameter values not set')

    def calc_background_no_thermo(self, params, do_reion=False):
        """
        Calculate the background evolution without calculating thermal or ionization history.
        e.g. call this if you want to just use :meth:`angular_diameter_distance` and similar background functions

        :param params:  :class:`~.model.CAMBparams` instance to use
        :param do_reion: whether to initialize the reionization model
        """
        self._check_params(params)
        self.f_SetParams(byref(params), None, byref(c_int(1 if do_reion else 0)), None, byref(c_int(1)))
        config.check_global_error('calc_background_no_thermo')

    def calc_background(self, params):
        """
        Calculate the background evolution and thermal history.
        e.g. call this if you want to get derived parameters and call background functions

        :param params:  :class:`~.model.CAMBparams` instance to use
        """
        self._check_params(params)
        if CAMBdata_CalcBackgroundTheory(byref(self), byref(params)):
            config.check_global_error('calc_background')

    def calc_transfers(self, params, only_transfers=True, only_time_sources=False):
        """
        Calculate the transfer functions (for CMB and matter power, as determined
        by params.WantCls, params.WantTransfer).

        :param params: :class:`~.model.CAMBparams` instance with parameters to use
        :param only_transfers: only calculate transfer functions, no power spectra
        :param only_time_sources: only calculate time transfer functions, no (p,l,k) transfer functions or
                                    non-linear scaling
        :return: non-zero if error, zero if OK
        """
        self._check_params(params)
        if not (only_transfers or only_time_sources):
            self._check_powers(params)
        if CAMBdata_gettransfers(byref(self), byref(params), byref(c_int(1 if only_transfers else 0)),
                                 byref(c_int(1 if only_time_sources else 0))):
            config.check_global_error('calc_transfer')

    def _check_powers(self, params=None):
        if params is None:
            params = self.Params
        if params.InitPower.has_tensors() and not params.WantTensors:
            raise CAMBError('r>0 but params.WantTensors = F')
        if params.WantScalars and params.WantCls and params.DoLensing and params.scalar_power(0.05) > 2e-8:
            raise CAMBError('Lensing requires a realistically normalized spectrum, you have P(k=0.05/Mpc) > 2e-8')

    def calc_power_spectra(self, params=None):
        """
        Calculates transfer functions and power spectra.

        :param params: optional :class:`~.model.CAMBparams` instance with parameters to use

        """
        if params is not None:
            self.calc_transfers(params, only_transfers=False)
        else:
            self._check_powers()
            CAMBdata_transferstopowers(byref(self))
            config.check_global_error()

    def power_spectra_from_transfer(self, initial_power_params=None, silent=False):
        """
        Assuming :meth:`calc_transfers` or :meth:`calc_power_spectra` have already been used, re-calculate the
        power spectra using a new set of initial power spectrum parameters with otherwise the same cosmology.
        This is typically much faster that re-calculating everything, as the transfer functions can be re-used.
        NOTE: if non-linear lensing is on, the transfer functions have the non-linear correction included when
        they are calculated, so using this function with a different initial power spectrum will not give quite the
        same results as doing a full recalculation.

        :param initial_power_params: :class:`.initialpower.InitialPowerLaw` or :class:`.initialpower.SplinedInitialPower`
               instance with new primordial power spectrum parameters, or None to use current power spectrum.
        :param silent: suppress warnings about non-linear corrections not being recalculated
        """
        if not silent and not self.HasScalarTimeSources and \
                self.Params.NonLinear in [model.NonLinear_lens, model.NonLinear_both] and \
                self.Params.WantScalars and self.Params.WantCls and not getattr(self, '_suppress_power_warn', False):
            logging.warning(
                'power_spectra_from_transfer with non-linear lensing does not recalculate the non-linear correction')
            self._suppress_power_warn = True
        if initial_power_params:
            self.Params.set_initial_power(initial_power_params)
        self._check_powers()
        CAMBdata_transferstopowers(byref(self))
        config.check_global_error()

    def _CMB_unit(self, CMB_unit):
        if isinstance(CMB_unit, str):
            if CMB_unit == 'muK':
                CMB_unit = self.Params.TCMB * 1e6
            elif CMB_unit == 'K':
                CMB_unit = self.Params.TCMB
            else:
                raise CAMBValueError('Unknown CMB_unit: %s' % CMB_unit)
        return CMB_unit

    def _scale_cls(self, cls, CMB_unit=None, raw_cl=False, lens_potential=False):
        if raw_cl:
            ls = np.arange(1, cls.shape[0])[..., np.newaxis]
            ls = np.float64(ls * (ls + 1))
            if lens_potential:
                cls[1:, 0] /= ls[:, 0] ** 2 / (2 * np.pi)
                cls[1:, 1:] /= ls ** (3. / 2) / (2 * np.pi)
            else:
                cls[1:, :] /= ls / (2 * np.pi)

        if CMB_unit is not None:
            CMB_unit = self._CMB_unit(CMB_unit)
            if lens_potential:
                cls[:, 1:] *= CMB_unit
            else:
                cls *= CMB_unit ** 2

        return cls

    def _lmax_setting(self, lmax=None, unlensed=False):
        if self.Params.DoLensing and not unlensed:
            lmax_calc = self.f_get_lmax_lensed()
            if not lmax_calc:
                raise CAMBError('lensed CL have not been calculated')
        else:
            lmax_calc = self.Params.max_l
        if lmax is None:
            lmax = lmax_calc
        elif lmax > lmax_calc:
            logging.warning('getting CMB power spectra to higher L than calculated, may be inaccurate/zeroed.')
        return lmax

    def save_cmb_power_spectra(self, filename, lmax=None, CMB_unit='muK'):
        r"""
        Save CMB power to a plain text file. Output is lensed total :math:`\ell(\ell+1)C_\ell/2\pi` then
        lensing potential and cross: L TT EE BB TE PP PT PE.

        :param filename: filename to save
        :param lmax: lmax to save
        :param CMB_unit: scale results from dimensionless. Use 'muK' for :math:`\mu K^2` units for CMB :math:`C_\ell`
               and :math:`\mu K` units for lensing cross.
        """
        lmax = self._lmax_setting(lmax)
        cmb = self.get_total_cls(lmax, CMB_unit=CMB_unit)
        lens = self.get_lens_potential_cls(lmax, CMB_unit=CMB_unit)
        save_cmb_power_array(filename, np.hstack((cmb, lens)), 'TT EE BB TE PP PT PE')

    def get_cmb_power_spectra(self, params=None, lmax=None,
                              spectra=('total', 'unlensed_scalar', 'unlensed_total', 'lensed_scalar', 'tensor',
                                       'lens_potential'), CMB_unit=None, raw_cl=False):
        r"""
        Get CMB power spectra, as requested by the 'spectra' argument. All power spectra are
        :math:`\ell(\ell+1)C_\ell/2\pi` self-owned numpy arrays (0..lmax, 0..3), where 0..3 index
        are TT, EE, BB, TE, unless raw_cl is True in which case return just :math:`C_\ell`.
        For the lens_potential the power spectrum returned is that of the deflection.

        Note that even if lmax is None, all spectra a returned to the same lmax, appropriate
        for lensed spectra. Use the individual functions instead if you want to the full unlensed
        and lensing potential power spectra to the higher lmax actually computed.

        :param params: optional :class:`~.model.CAMBparams` instance with parameters to use. If None, must have
          previously set parameters and called `calc_power_spectra` (e.g. if you got this instance
          using :func:`.camb.get_results`),
        :param lmax: maximum L
        :param spectra: list of names of spectra to get
        :param CMB_unit: scale results from dimensionless. Use 'muK' for :math:`\mu K^2` units for CMB :math:`C_\ell`
          and :math:`\mu K` units for lensing cross.
        :param raw_cl: return :math:`C_\ell` rather than :math:`\ell(\ell+1)C_\ell/2\pi`
        :return: dictionary of power spectrum arrays, indexed by names of requested spectra
        """
        P = {}
        if params is not None:
            self.calc_power_spectra(params)
        lmax = self._lmax_setting(lmax)
        for spectrum in spectra:
            P[spectrum] = getattr(self, 'get_' + spectrum + '_cls')(lmax, CMB_unit=CMB_unit,
                                                                    raw_cl=raw_cl)
        return P

    def get_cmb_correlation_functions(self, params=None, lmax=None, spectrum='lensed_scalar',
                                      xvals=None, sampling_factor=1):
        r"""
        Get the CMB correlation functions from the power spectra.
        By default evaluated at points :math:`\cos(\theta)` = xvals that are roots of Legendre polynomials,
        for accurate back integration with :func:`.correlations.corr2cl`.
        If xvals is explicitly given, instead calculates correlations at provided :math:`\cos(\theta)` values.

        :param params: optional :class:`~.model.CAMBparams` instance with parameters to use. If None, must have
          previously set parameters and called :meth:`calc_power_spectra` (e.g. if you got this instance
          using :func:`.camb.get_results`),
        :param lmax: optional maximum L to use from the cls arrays
        :param spectrum: type of CMB power spectrum to get; default 'lensed_scalar', one of
          ['total', 'unlensed_scalar', 'unlensed_total', 'lensed_scalar', 'tensor']
        :param xvals: optional array of :math:`\cos(\theta)` values at which to calculate correlation function.
        :param sampling_factor: multiple of lmax for the Gauss-Legendre order if xvals not given (default 1)
        :return: if xvals not given: corrs, xvals, weights; if xvals specified, just corrs.
          corrs is 2D array corrs[i, ix], where ix=0,1,2,3 are T, Q+U, Q-U and cross, and i indexes xvals
        """

        if spectrum not in ['total', 'unlensed_scalar', 'unlensed_total', 'lensed_scalar', 'tensor']:
            raise CAMBValueError('Can only get CMB correlation functions for known CMB spectrum')
        from . import correlations

        cls = self.get_cmb_power_spectra(params, lmax, spectra=[spectrum])[spectrum]
        if xvals is None:
            return correlations.gauss_legendre_correlation(cls, sampling_factor=sampling_factor)
        else:
            return correlations.cl2corr(cls, xvals, lmax=lmax)

    def get_cmb_transfer_data(self, tp='scalar'):
        r"""
        Get :math:`C_\ell` transfer functions

        :return: :class:`.ClTransferData` instance holding output arrays (copies, not pointers)
        """

        cdata = _ClTransferData()
        CAMBdata_cltransferdata(byref(self), byref(cdata),
                                byref(c_int(['scalar', 'vector', 'tensor'].index(tp))))
        data = ClTransferData()
        data.NumSources = cdata.NumSources
        data.q = fortran_array(cdata.q, cdata.q_size)
        data.L = fortran_array(cdata.L, cdata.l_size, dtype=c_int)
        data.delta_p_l_k = fortran_array(cdata.delta_p_l_k, cdata.delta_size)
        return data

    def get_time_evolution(self, q, eta, vars=model.evolve_names, lAccuracyBoost=4, frame='CDM'):
        """
        Get the mode evolution as a function of conformal time for some k values.

        Note that gravitational potentials (e.g. Weyl) are not integrated in the code and are
        calculated as derived parameters; they may be numerically unstable far outside the horizon.
        (use the series expansion result if needed far outside the horizon)

        :param q: wavenumber values to calculate (or array of k values)
        :param eta: array of requested conformal times to output
        :param vars: list of variable names or sympy symbolic expressions to output (using camb.symbolic)
        :param lAccuracyBoost: factor by which to increase l_max in hierarchies compared to default - often
          needed to get nice smooth curves of acoustic oscillations for plotting.
        :param frame: for symbolic expressions, can specify frame name if the variable is not gauge invariant.
            e.g. specifying Delta_g and frame='Newtonian' would give the Newtonian gauge photon density perturbation.
        :return: nd array, A_{qti}, size(q) x size(times) x len(vars), or 2d array if q is scalar
        """

        old_boost = self.Params.Accuracy.lAccuracyBoost
        try:
            if lAccuracyBoost:
                self.Params.Accuracy.lAccuracyBoost = lAccuracyBoost
            if not isinstance(vars, (tuple, list)):
                vars = [vars]
            import sympy
            named_vars = [var for var in vars if isinstance(var, str)]

            unknown = set(named_vars) - set(model.evolve_names)
            if unknown:
                raise CAMBError('Unknown names %s; valid names are %s' % (unknown, model.evolve_names))

            num_standard_names = len(model.evolve_names)

            custom_vars = []
            ix = np.empty(len(vars), dtype=int)
            for i, var in enumerate(vars):
                if var in model.evolve_names:
                    ix[i] = model.evolve_names.index(var)
                elif isinstance(var, sympy.Expr):
                    custom_vars.append(var)
                    ix[i] = num_standard_names + len(custom_vars) - 1
                else:
                    raise CAMBError(
                        'Variables must be variable names, or a sympy expression (using camb.symbolic variables)')

            if np.isscalar(q):
                k = np.array([q], dtype=np.float64)
            else:
                k = np.array(q, dtype=np.float64)
            times = np.array(np.atleast_1d(eta), dtype=np.float64)
            indices = np.argsort(times)  # times must be in increasing order
            ncustom = len(custom_vars)
            if ncustom:
                from . import symbolic
                funcPtr = symbolic.compile_sympy_to_camb_source_func(custom_vars, frame=frame)
                custom_source_func = ctypes.cast(funcPtr, ctypes.c_void_p)
            else:
                custom_source_func = ctypes.c_void_p(0)
            nvars = num_standard_names + ncustom
            outputs = np.empty((k.shape[0], times.shape[0], nvars))
            if CAMB_TimeEvolution(byref(self), byref(c_int(k.shape[0])), k, byref(c_int(times.shape[0])),
                                  times[indices], byref(c_int(nvars)), outputs,
                                  byref(c_int(ncustom)), byref(custom_source_func)):
                config.check_global_error('get_time_evolution')
            i_rev = np.zeros(times.shape, dtype=int)
            i_rev[indices] = np.arange(times.shape[0])
            outputs = outputs[:, i_rev, :]
        finally:
            self.Params.Accuracy.lAccuracyBoost = old_boost
        if np.isscalar(q):
            return outputs[0, :, :][:, ix]
        else:
            return outputs[:, :, ix]

    def get_redshift_evolution(self, q, z, vars=model.evolve_names, lAccuracyBoost=4):
        """
        Get the mode evolution as a function of redshift for some k values.

        :param q: wavenumber values to calculate (or array of k values)
        :param z: array of redshifts to output
        :param vars: list of variable names or camb.symbolic sympy expressions to output
        :param lAccuracyBoost: boost factor for ell accuracy (e.g. to get nice smooth curves for plotting)
        :return: nd array, A_{qti}, size(q) x size(times) x len(vars), or 2d array if q is scalar
        """
        return self.get_time_evolution(q, self.conformal_time(z), vars, lAccuracyBoost)

    def get_background_time_evolution(self, eta, vars=model.background_names, format='dict'):
        """
        Get the evolution of background variables a function of conformal time.
        For the moment a and H are rather perversely only available via :meth:`get_time_evolution`

        :param eta: array of requested conformal times to output
        :param vars: list of variable names to output
        :param format: 'dict' or 'array', for either dict of 1D arrays indexed by name, or 2D array
        :return: n_eta x len(vars) 2D numpy array of outputs or dict of 1D arrays
        """

        if isinstance(vars, str):
            vars = [vars]
        unknown = set(vars) - set(model.background_names)
        if unknown:
            raise CAMBError('Unknown names %s; valid names are %s' % (unknown, model.background_names))
        outputs = np.zeros((eta.shape[0], 9))
        CAMB_BackgroundThermalEvolution(byref(self), byref(c_int(eta.shape[0])), eta, outputs)
        indices = [model.background_names.index(var) for var in vars]
        if format == 'dict':
            res = {}
            for var, index in zip(vars, indices):
                res[var] = outputs[:, index]
            return res
        else:
            assert format == 'array', "format must be dict or array"
            return outputs[:, np.array(indices)]

    def get_background_redshift_evolution(self, z, vars=model.background_names, format='dict'):
        """
        Get the evolution of background variables a function of redshift.
        For the moment a and H are rather perversely only available via :meth:`get_time_evolution`

        :param z: array of requested redshifts to output
        :param vars: list of variable names to output
        :param format: 'dict' or 'array', for either dict of 1D arrays indexed by name, or 2D array
        :return: n_eta x len(vars) 2D numpy array of outputs or dict of 1D arrays
        """

        return self.get_background_time_evolution(self.conformal_time(z), vars, format)

    def get_background_densities(self, a, vars=model.density_names, format='dict'):
        r"""
        Get the individual densities as a function of scale factor. Returns :math:`8\pi G a^4 \rho_i` in Mpc units.
        :math:`\Omega_i` can be simply obtained by taking the ratio of the components to tot.

        :param a: scale factor or array of scale factors
        :param vars: list of variables to output (default all)
        :param format: 'dict' or 'array', for either dict of 1D arrays indexed by name, or 2D array
        :return: n_a x len(vars) 2D numpy array or dict of 1D arrays of :math:`8\pi G a^4 \rho_i` in Mpc units.
        """
        if isinstance(vars, str):
            vars = [vars]
        unknown = set(vars) - set(model.density_names)
        if unknown:
            raise CAMBError('Unknown names %s; valid names are %s' % (unknown, model.density_names))
        arr = np.atleast_1d(a)
        outputs = np.zeros((arr.shape[0], 8))
        self.f_GetBackgroundDensities(byref(c_int(arr.shape[0])), arr, outputs)
        indices = [model.density_names.index(var) for var in vars]
        if format == 'dict':
            res = {}
            for var, index in zip(vars, indices):
                res[var] = outputs[:, index]
            return res
        else:
            assert format == 'array', "format must be dict or array"
            return outputs[:, np.array(indices)]

    def get_dark_energy_rho_w(self, a):
        r"""
        Get dark energy density in units of the dark energy density today, and equation of state parameter
        :math:`w\equiv P/\rho`

        :param a: scalar factor or array of scale factors
        :return: rho, w arrays at redshifts :math:`1/a-1` [or scalars if :math:`a` is scalar]
        """
        if np.isscalar(a):
            scales = np.array([a])
        else:
            scales = np.ascontiguousarray(a)
        rho = np.zeros(scales.shape)
        w = np.zeros(scales.shape)
        self.f_DarkEnergyStressEnergy(scales, rho, w, byref(c_int(len(scales))))
        if np.isscalar(a):
            return rho[0], w[0]
        else:
            return rho, w

    def get_Omega(self, var, z=0):
        r"""
        Get density relative to critical density of variables var

        :param var: one of 'K', 'cdm', 'baryon', 'photon', 'neutrino' (massless), 'nu' (massive neutrinos), 'de'
        :param z: redshift
        :return:  :math:`\Omega_i(a)`
        """
        dic = self.get_background_densities(1. / (1 + z), ['tot', var])
        res = dic[var] / dic['tot']
        if np.isscalar(z):
            return res[0]
        else:
            return res

    def get_matter_transfer_data(self) -> MatterTransferData:
        """
        Get matter transfer function data and sigma8 for calculated results.

        :return: :class:`.MatterTransferData` instance holding output arrays (copies, not pointers)
        """
        if not self.Params.WantTransfer:
            raise CAMBError("must have Params.WantTransfer to get matter transfers and power")

        cdata = _MatterTransferData()
        CAMBdata_mattertransferdata(byref(self), byref(cdata))
        data = MatterTransferData()
        data.nq = cdata.num_q_trans
        from numpy import ctypeslib as nplib
        data.q = nplib.as_array(cdata.q_trans, shape=(data.nq,)).copy()
        if cdata.sigma_8_size:
            data.sigma_8 = nplib.as_array(cdata.sigma_8, shape=(cdata.sigma_8_size,)).copy()
        else:
            data.sigma_8 = None
        if cdata.sigma2_vdelta_8_size:
            data.sigma2_vdelta_8 = nplib.as_array(cdata.sigma2_vdelta_8, shape=(cdata.sigma2_vdelta_8_size,)).copy()
        else:
            data.sigma2_vdelta_8 = None
        data.transfer_data = fortran_array(cdata.TransferData, cdata.TransferData_size, dtype=np.float32)
        return data

    @staticmethod
    def _transfer_var(var1, var2):
        if var1 is None:
            var1 = config.transfer_power_var
        if var2 is None:
            var2 = config.transfer_power_var
        if isinstance(var1, str):
            var1 = model.transfer_names.index(var1) + 1
        if isinstance(var2, str):
            var2 = model.transfer_names.index(var2) + 1
        return c_int(var1), c_int(var2)

    def get_linear_matter_power_spectrum(self, var1=None, var2=None, hubble_units=True, k_hunit=True,
                                         have_power_spectra=True, params=None, nonlinear=False):
        r"""
        Calculates :math:`P_{xy}(k)`, where x, y are one of model.Transfer_cdm, model.Transfer_xx etc.
        The output k values are not regularly spaced, and not interpolated. They are either k or k/h depending on the
        value of k_hunit (default True gives k/h).

        For a description of outputs for different var1, var2 see :ref:`transfer-variables`.

        :param var1: variable i (index, or name of variable; default delta_tot)
        :param var2: variable j (index, or name of variable; default delta_tot)
        :param hubble_units: if true, output power spectrum in (Mpc/h) units, otherwise Mpc
        :param k_hunit: if true, matter power is a function of k/h, if false, just k (both :math:`{\rm Mpc}^{-1}` units)
        :param have_power_spectra: set to False if not already computed power spectra
        :param params: if have_power_spectra=False, optional :class:`~.model.CAMBparams` instance
                       to specify new parameters
        :param nonlinear: include non-linear correction from halo model
        :return: k/h or k, z, PK, where kz and z are arrays of k/h or k and z respectively,
                 and PK[i,j] is the value at z[i], k[j]/h or k[j]
        """
        if self.OnlyTransfers or params is not None or not have_power_spectra:
            self.calc_power_spectra(params)

        num_k = c_int(0)
        CAMBdata_mattertransferks(byref(self), byref(num_k), np.array([]))
        nk = num_k.value

        ks = np.empty(nk, dtype=np.float64)
        CAMBdata_mattertransferks(byref(self), byref(num_k), ks)

        nz = self.Params.Transfer.PK_num_redshifts
        if k_hunit:
            kh = ks / (self.Params.H0 / 100)
        else:
            kh = ks
        var1, var2 = self._transfer_var(var1, var2)

        hubble_units = c_int(hubble_units)
        PK = np.empty((nz, nk))
        if nonlinear:
            CAMBdata_GetNonLinearMatterPower(byref(self), PK, byref(var1), byref(var2), byref(hubble_units))
            config.check_global_error('get_[non]linear_matter_power_spectrum')
        else:
            CAMBdata_GetLinearMatterPower(byref(self), PK, byref(var1), byref(var2), byref(hubble_units))

        z = self.Params.Transfer.PK_redshifts[:nz]
        z.reverse()
        return kh, np.array(z), PK

    def get_nonlinear_matter_power_spectrum(self, var1=None, var2=None, hubble_units=True, k_hunit=True,
                                            have_power_spectra=True, params=None):
        r"""
        Calculates :math:`P_{xy}(k/h)`, where x, y are one of model.Transfer_cdm, model.Transfer_xx etc.
        The output k values are not regularly spaced, and not interpolated.

        For a description of outputs for different var1, var2 see :ref:`transfer-variables`.

        :param var1: variable i (index, or name of variable; default delta_tot)
        :param var2: variable j (index, or name of variable; default delta_tot)
        :param hubble_units: if true, output power spectrum in :Math:`({\rm Mpc}/h)^{3}` units,
                             otherwise :math:`{\rm Mpc}^{3}`
        :param k_hunit: if true, matter power is a function of k/h, if false, just k (both :math:`{\rm Mpc}^{-1}` units)
        :param have_power_spectra: set to False if not already computed power spectra
        :param params: if have_power_spectra=False, optional :class:`~.model.CAMBparams` instance
                       to specify new parameters
        :return: k/h or k, z, PK, where kz and z are arrays of k/h or k and z respectively,
                 and PK[i,j] is the value at z[i], k[j]/h or k[j]
        """
        return self.get_linear_matter_power_spectrum(var1=var1, var2=var2, hubble_units=hubble_units, k_hunit=k_hunit,
                                                     have_power_spectra=have_power_spectra, params=params,
                                                     nonlinear=True)

    def get_sigmaR(self, R, z_indices=None, var1=None, var2=None, hubble_units=True, return_R_z=False):
        r"""
        Calculate :math:`\sigma_R` values, the RMS linear matter fluctuation in spheres of radius R in linear theory.
        Accuracy depends on the sampling with which the matter transfer functions are computed.

        For a description of outputs for different var1, var2 see :ref:`transfer-variables`.
        Note that numerical errors are slightly different to get_sigma8 for R=8 Mpc/h.

        :param R: radius in Mpc or h^{-1} Mpc units, scalar or array
        :param z_indices: indices of redshifts in Params.Transfer.PK_redshifts to calculate
                          (default None gives all computed in order of increasing time (decreasing redshift);
                          -1 gives redshift 0; list gives all listed indices)
        :param var1: variable i (index, or name of variable; default delta_tot)
        :param var2: variable j (index, or name of variable; default delta_tot)
        :param hubble_units: if true, R is in h^{-1} Mpc, otherwise Mpc
        :param return_R_z: if true, return tuple of R, z, sigmaR
                           (where R always Mpc units not h^{-1}Mpc and R, z are arrays)
        :return: array of :math:`\sigma_R` values, or 2D array indexed by (redshift, R)
        """
        assert self.Params.WantTransfer
        var1, var2 = self._transfer_var(var1, var2)
        if hubble_units:
            if not np.isscalar(R):
                R = np.array(R, dtype=np.float64)
            R = R / (self.Params.H0 / 100)

        radii = np.atleast_1d(np.ascontiguousarray(R, dtype=np.float64)) * self.Params.H0 / 100
        valid_indices = np.array(self.PK_redshifts_index[:self.Params.Transfer.PK_num_redshifts], dtype=np.int32)
        if z_indices is None:
            z_ix = valid_indices
        else:
            z_ix = np.atleast_1d(valid_indices[z_indices])
        sigma_R = np.empty((len(z_ix), len(radii)), dtype=np.float64)
        CAMBdata_GetSigmaRArray(byref(self), sigma_R, radii, byref(c_int(len(radii))), z_ix,
                                byref(c_int(len(z_ix))), byref(var1), byref(var2))
        if z_indices is not None and np.isscalar(z_indices):
            sigma_R = sigma_R[0, :]
            if np.isscalar(R):
                sigma_R = sigma_R[0]
        elif np.isscalar(R):
            sigma_R = sigma_R[:, 0]
        if return_R_z:
            return R, np.array(self.transfer_redshifts)[z_ix - 1], sigma_R
        else:
            return sigma_R

    def get_sigma8(self):
        r"""
        Get :math:`\sigma_8` values at Params.PK_redshifts (must previously have calculated power spectra)

        :return: array of :math:`\sigma_8` values, in order of increasing time (decreasing redshift)
        """
        assert self.Params.WantTransfer
        sigma8 = np.empty(self.Params.Transfer.PK_num_redshifts, dtype=np.float64)
        CAMBdata_GetSigma8(byref(self), sigma8, byref(c_int(0)))
        return sigma8

    def get_sigma8_0(self):
        r"""
        Get :math:`\sigma_8` value today (must previously have calculated power spectra)

        :return: :math:`\sigma_8` today
        """
        nz = self.Params.Transfer.PK_num_redshifts
        if not nz or self.Params.Transfer.PK_redshifts[nz - 1] > 1e-5:
            raise CAMBError("sigma8 requested at z=0, but P(z=0) not calculated")
        return self.get_sigma8()[-1]

    def get_fsigma8(self):
        r"""
        Get :math:`f\sigma_8` growth values (must previously have calculated power spectra).
        For general models :math:`f\sigma_8` is defined as in the Planck 2015 parameter paper in terms of
        the velocity-density correlation: :math:`\sigma^2_{vd}/\sigma_{dd}` for :math:`8 h^{-1} {\rm Mpc}` spheres.

        :return: array of f*sigma_8 values, in order of increasing time (decreasing redshift)
        """
        assert self.Params.WantTransfer
        fsigma8 = np.empty(self.Params.Transfer.PK_num_redshifts, dtype=np.float64)
        CAMBdata_GetSigma8(byref(self), fsigma8, byref(c_int(1)))
        return fsigma8

    def get_matter_power_spectrum(self, minkh=1e-4, maxkh=1.0, npoints=100,
                                  var1=None, var2=None,
                                  have_power_spectra=False, params=None):
        """
        Calculates :math:`P_{xy}(k/h)`, where x, y are one of Transfer_cdm, Transfer_xx etc.
        The output k values are regularly log spaced and interpolated. If NonLinear is set, the result is non-linear.

        For a description of outputs for different var1, var2 see :ref:`transfer-variables`.

        :param minkh: minimum value of k/h for output grid (very low values < 1e-4 may not be calculated)
        :param maxkh: maximum value of k/h (check consistent with input params.Transfer.kmax)
        :param npoints: number of points equally spaced in log k
        :param var1: variable i (index, or name of variable; default delta_tot)
        :param var2: variable j (index, or name of variable; default delta_tot)
        :param have_power_spectra: set to True if already computed power spectra
        :param params: if have_power_spectra=False and want to specify new parameters,
                       a :class:`~.model.CAMBparams` instance
        :return: kh, z, PK, where kz and z are arrays of k/h and z respectively, and PK[i,j] is value at z[i], k/h[j]
        """

        if not have_power_spectra:
            self.calc_power_spectra(params)

        if not npoints >= 2:
            raise CAMBError('Need at least two points in get_matter_power_spectrum')

        assert self.Params.WantTransfer
        if self.Params.Transfer.kmax < maxkh * self.Params.h:
            logging.warning("get_matter_power_spectrum using larger k_max than input parameter Transfer.kmax")
        if self.Params.NonLinear != model.NonLinear_none and self.Params.Transfer.kmax < 1:
            logging.warning("get_matter_power_spectrum Transfer.kmax small to get non-linear spectrum")

        nz = self.Params.Transfer.PK_num_redshifts
        PK = np.empty((nz, npoints))
        var1, var2 = self._transfer_var(var1, var2)

        dlnkh = (np.log(maxkh) - np.log(minkh)) / (npoints - 1)
        CAMBdata_GetMatterPower(byref(self), PK, byref(c_double(minkh)),
                                byref(c_double(dlnkh)), byref(c_int(npoints)), byref(var1), byref(var2))
        z = self.Params.Transfer.PK_redshifts[:nz]
        z.reverse()
        return minkh * np.exp(np.arange(npoints) * dlnkh), z, PK

    def get_matter_power_interpolator(self, nonlinear=True, var1=None, var2=None, hubble_units=True, k_hunit=True,
                                      return_z_k=False, log_interp=True, extrap_kmax=None, silent=False):
        r"""
        Assuming transfers have been calculated, return a 2D spline interpolation object to evaluate matter
        power spectrum as function of z and k/h (or k). Uses self.Params.Transfer.PK_redshifts as the spline node
        points in z. If fewer than four redshift points are used the interpolator uses a reduced order spline in z
        (so results at intermediate z may be inaccurate), otherwise it uses bicubic.
        Usage example:

        .. code-block:: python

           PK = results.get_matter_power_interpolator();
           print('Power spectrum at z=0.5, k/h=0.1 is %s (Mpc/h)^3 '%(PK.P(0.5, 0.1)))

        For a description of outputs for different var1, var2 see :ref:`transfer-variables`.

        :param nonlinear: include non-linear correction from halo model
        :param var1: variable i (index, or name of variable; default delta_tot)
        :param var2: variable j (index, or name of variable; default delta_tot)
        :param hubble_units: if true, output power spectrum in :math:`({\rm Mpc}/h)^{3}` units,
                             otherwise :math:`{\rm Mpc}^{3}`
        :param k_hunit: if true, matter power is a function of k/h, if false, just k (both :math:`{\rm Mpc}^{-1}` units)
        :param return_z_k: if true, return interpolator, z, k where z, k are the grid used
        :param log_interp: if true, interpolate log of power spectrum
                           (unless any values cross zero in which case ignored)
        :param extrap_kmax: if set, use power law extrapolation beyond kmax to extrap_kmax
                            (useful for tails of integrals)
        :param silent: Set True to silence warnings
        :return: An object PK based on :class:`~scipy:scipy.interpolate.RectBivariateSpline`,
                 that can be called with PK.P(z,kh) or PK(z,log(kh)) to get log matter power values.
                 If return_z_k=True, instead return interpolator, z, k where z, k are the grid used.
        """

        class PKInterpolator(RectBivariateSpline):
            islog: bool
            logsign: int

            def P(self, z, kh, grid=None):
                if grid is None:
                    grid = not np.isscalar(z) and not np.isscalar(kh)
                if self.islog:
                    return self.logsign * np.exp(self(z, np.log(kh), grid=grid))
                else:
                    return self(z, np.log(kh), grid=grid)

        class PKInterpolatorSingleZ(interp1d):
            islog: bool
            logsign: int

            def __init__(self, *args, **kwargs):
                self._single_z = np.array(args[0])
                super().__init__(*(args[1:]), kind=kwargs.get("ky"))

            def check_z(self, z):
                if not np.allclose(z, self._single_z):
                    raise CAMBError(
                        "P(z,k) requested at z=%g, but only computed for z=%s. "
                        "Cannot extrapolate!" % (z, self._single_z))

            def __call__(self, *args):
                self.check_z(args[0])
                # NB returns dimensionality as the 2D one: 1 dimension if z single
                return (lambda x: x[0] if np.isscalar(args[0]) else x)(super().__call__(*(args[1:])))

            def P(self, z, kh, **_kwargs):
                # grid kwarg is ignored
                if self.islog:
                    return self.logsign * np.exp(self(z, np.log(kh)))
                else:
                    return self(z, np.log(kh))

        assert self.Params.WantTransfer
        khs, zs, pk = self.get_linear_matter_power_spectrum(var1, var2, hubble_units, nonlinear=nonlinear)
        kh_max = khs[-1]
        if not k_hunit:
            khs *= self.Params.H0 / 100
        sign = 1
        if log_interp and np.any(pk <= 0):
            if np.all(pk < 0):
                sign = -1
            else:
                log_interp = False
        p_or_log_p = np.log(sign * pk) if log_interp else pk
        logkh = np.log(khs)
        deg_z = min(len(zs) - 1, 3)
        kmax = khs[-1]
        PKInterpolator = PKInterpolator if deg_z else PKInterpolatorSingleZ
        if extrap_kmax and extrap_kmax > kmax:
            # extrapolate to ultimate power law
            # TODO: use more physical extrapolation function for linear case
            if not silent and (kh_max < 3 and extrap_kmax > 2 and nonlinear or kh_max < 0.4):
                logging.warning("Extrapolating to higher k with matter transfer functions "
                                "only to k=%.3g Mpc^{-1} may be inaccurate.\n " % (kh_max * self.Params.H0 / 100))
            if not log_interp:
                raise CAMBValueError(
                    "Cannot use extrap_kmax with log_inter=False (e.g. PK crossing zero for %s, %s.)" % (var1, var2))

            logextrap = np.log(extrap_kmax)
            log_p_new = np.empty((pk.shape[0], pk.shape[1] + 2))
            log_p_new[:, :-2] = p_or_log_p
            delta = logextrap - logkh[-1]

            dlog = (log_p_new[:, -3] - log_p_new[:, -4]) / (logkh[-1] - logkh[-2])
            log_p_new[:, -1] = log_p_new[:, -3] + dlog * delta
            log_p_new[:, -2] = log_p_new[:, -3] + dlog * delta * 0.9
            logkh = np.hstack((logkh, logextrap - delta * 0.1, logextrap))
            p_or_log_p = log_p_new

        deg_k = min(len(logkh) - 1, 3)
        res = PKInterpolator(zs, logkh, p_or_log_p, kx=deg_z, ky=deg_k)
        res.kmin = np.min(khs)
        res.kmax = kmax
        res.islog = log_interp
        res.logsign = sign
        res.zmin = np.min(zs)
        res.zmax = np.max(zs)
        if return_z_k:
            return res, zs, khs
        else:
            return res

    def get_total_cls(self, lmax=None, CMB_unit=None, raw_cl=False):
        r"""
        Get lensed-scalar + tensor CMB power spectra. Must have already calculated power spectra.

        :param lmax: lmax to output to
        :param CMB_unit: scale results from dimensionless. Use 'muK' for :math:`\mu K^2` units for CMB :math:`C_\ell`
        :param raw_cl: return :math:`C_\ell` rather than :math:`\ell(\ell+1)C_\ell/2\pi`
        :return: numpy array CL[0:lmax+1,0:4], where 0..3 indexes TT, EE, BB, TE
        """
        lmax = self._lmax_setting(lmax)
        res = np.empty((lmax + 1, 4))
        opt = c_int(lmax)
        CAMB_SetTotCls(byref(self), byref(opt), res)
        self._scale_cls(res, CMB_unit, raw_cl)
        return res

    def get_tensor_cls(self, lmax=None, CMB_unit=None, raw_cl=False):
        r"""
        Get tensor CMB power spectra. Must have already calculated power spectra.

        :param lmax: lmax to output to
        :param CMB_unit: scale results from dimensionless. Use 'muK' for :math:`\mu K^2` units for CMB :math:`C_\ell`
        :param raw_cl: return :math:`C_\ell` rather than :math:`\ell(\ell+1)C_\ell/2\pi`
        :return: numpy array CL[0:lmax+1,0:4], where 0..3 indexes TT, EE, BB, TE
        """

        if lmax is None:
            lmax = self.Params.max_l_tensor
        lmax = self._lmax_setting(lmax, unlensed=True)
        res = np.empty((lmax + 1, 4))
        opt = c_int(lmax)
        CAMB_SetTensorCls(byref(self), byref(opt), res)
        self._scale_cls(res, CMB_unit, raw_cl)
        return res

    def get_unlensed_scalar_cls(self, lmax=None, CMB_unit=None, raw_cl=False):
        r"""
        Get unlensed scalar CMB power spectra. Must have already calculated power spectra.

        :param lmax: lmax to output to
        :param CMB_unit: scale results from dimensionless. Use 'muK' for :math:`\mu K^2` units for CMB :math:`C_\ell`
        :param raw_cl: return :math:`C_\ell` rather than :math:`\ell(\ell+1)C_\ell/2\pi`
        :return: numpy array CL[0:lmax+1,0:4], where 0..3 indexes TT, EE, BB, TE. CL[:,2] will be zero.
        """
        lmax = self._lmax_setting(lmax, unlensed=True)
        res = np.empty((lmax + 1, 4))
        opt = c_int(lmax)
        CAMB_SetUnlensedScalCls(byref(self), byref(opt), res)
        self._scale_cls(res, CMB_unit, raw_cl)
        return res

    def get_unlensed_total_cls(self, lmax=None, CMB_unit=None, raw_cl=False):
        r"""
        Get unlensed CMB power spectra, including tensors if relevant. Must have already calculated power spectra.

        :param lmax: lmax to output to
        :param CMB_unit: scale results from dimensionless. Use 'muK' for :math:`\mu K^2` units for CMB :math:`C_\ell`
        :param raw_cl: return :math:`C_\ell` rather than :math:`\ell(\ell+1)C_\ell/2\pi`
        :return: numpy array CL[0:lmax+1,0:4], where 0..3 indexes TT, EE, BB, TE.
        """
        lmax = self._lmax_setting(lmax, unlensed=True)
        return self.get_unlensed_scalar_cls(lmax, CMB_unit, raw_cl) + self.get_tensor_cls(lmax, CMB_unit, raw_cl)

    def get_lensed_scalar_cls(self, lmax=None, CMB_unit=None, raw_cl=False):
        r"""
        Get lensed scalar CMB power spectra. Must have already calculated power spectra.

        :param lmax: lmax to output to
        :param CMB_unit: scale results from dimensionless. Use 'muK' for :math:`\mu K^2` units for CMB :math:`C_\ell`
        :param raw_cl: return :math:`C_\ell` rather than :math:`\ell(\ell+1)C_\ell/2\pi`
        :return: numpy array CL[0:lmax+1,0:4], where 0..3 indexes TT, EE, BB, TE.
        """

        lmax = self._lmax_setting(lmax)
        res = np.empty((lmax + 1, 4))
        opt = c_int(lmax)
        CAMB_SetLensedScalCls(byref(self), byref(opt), res)
        self._scale_cls(res, CMB_unit, raw_cl)
        return res

    def get_lens_potential_cls(self, lmax=None, CMB_unit=None, raw_cl=False):
        r"""
        Get lensing deflection angle potential power spectrum, and cross-correlation with T and E. Must have already
        calculated power spectra.
        Power spectra are :math:`[L(L+1)]^2C_L^{\phi\phi}/2\pi` and corresponding deflection cross-correlations.

        :param lmax: lmax to output to

        :param CMB_unit: scale results from dimensionless. Use 'muK' for :math:`\mu K` units for lensing cross.
        :param raw_cl: return lensing potential :math:`C_L` rather than :math:`[L(L+1)]^2C_L/2\pi`
        :return: numpy array CL[0:lmax+1,0:3], where 0..2 indexes PP, PT, PE.
        """

        lmax = self._lmax_setting(lmax, unlensed=True)
        res = np.empty((lmax + 1, 3))
        opt = c_int(lmax)
        CAMB_SetLensPotentialCls(byref(self), byref(opt), res)
        self._scale_cls(res, CMB_unit, raw_cl, lens_potential=True)
        return res

    def get_unlensed_scalar_array_cls(self, lmax=None):
        r"""
        Get array of all cross power spectra. Must have already calculated power spectra.
        Results are dimensionless, and not scaled by custom_scaled_ell_fac.

        :param lmax: lmax to output to
        :return: numpy array CL[0:, 0:,0:lmax+1], where 0.. index T, E, lensing potential, source window functions
        """

        lmax = self._lmax_setting(lmax, unlensed=True)
        if not self.Params.Want_cl_2D_array:
            raise CAMBError('unlensed_scalar_array not calculated (set Want_cl_2D_array)')

        n = 3 + len(self.Params.SourceWindows) + self.Params.CustomSources.num_custom_sources
        res = np.empty((n, n, lmax + 1), order='F')
        CAMB_SetUnlensedScalarArray(byref(self), byref(c_int(lmax)), res, byref(c_int(n)))
        return res

    def get_cmb_unlensed_scalar_array_dict(self, params=None, lmax=None, CMB_unit=None, raw_cl=False):
        r"""
        Get all unlensed auto and cross power spectra, including any custom source functions set
        using :meth:`.model.CAMBparams.set_custom_scalar_sources`.

        :param params: optional :class:`~.model.CAMBparams` instance with parameters to use. If None, must have
          previously set parameters and called :meth:`calc_power_spectra` (e.g. if you got this instance
          using :func:`.camb.get_results`),
        :param lmax: maximum :math:`\ell`
        :param CMB_unit: scale results from dimensionless. Use 'muK' for :math:`\mu K^2` units for
          CMB :math:`C_\ell` and :math:`\mu K` units for lensing cross.
        :param raw_cl: return :math:`C_\ell` rather than :math:`\ell(\ell+1)C_\ell/2\pi`
        :return: dictionary of power spectrum arrays, index as TxT, TxE, PxW1, W1xW2, custom_name_1xT... etc.
                 Note that P is the lensing deflection, lensing windows Wx give convergence.
        """

        old_val = None
        try:
            if params is not None:
                old_val = params.Want_cl_2D_array
                params.Want_cl_2D_array = True
                self.calc_power_spectra(params)
            elif not self.Params.Want_cl_2D_array:
                raise CAMBValueError('Want_cl_2D_array must be true to have array C_L')
            nwindows = len(self.Params.SourceWindows)
            lmax = lmax or self.Params.max_l
            arr = self.get_unlensed_scalar_array_cls(lmax)
            custom_source_names = self.Params.get_custom_source_names()
            names = ['T', 'E', 'P'] + ["W%s" % (i + 1) for
                                       i in range(nwindows)] + custom_source_names
            CMB_unit = self._CMB_unit(CMB_unit) or 1
            CMB_units = [CMB_unit, CMB_unit, 1] + [1] * nwindows + [CMB_unit] * len(custom_source_names)

            result = {}
            for i, name in enumerate(names):
                for j, name2 in enumerate(names):
                    tag = name + 'x' + name2
                    if j < i:
                        result[tag] = result[name2 + 'x' + name]
                    else:
                        cls = arr[i, j, :]
                        if raw_cl:
                            ls = np.arange(1, cls.shape[0])
                            fac = np.float64(ls * (ls + 1))
                            if i == 2 and j == 2:
                                fac *= fac
                            elif i == 2 or j == 2:
                                fac *= np.sqrt(fac)
                            cls[1:] /= (fac / (2 * np.pi))

                        if CMB_unit is not None:
                            cls *= CMB_units[i] * CMB_units[j]
                        result[tag] = cls
        finally:
            if old_val is not None:
                params.Want_cl_2D_array = old_val
        return result

    def get_source_cls_dict(self, params=None, lmax=None, raw_cl=False):
        r"""
        Get all source window function and CMB lensing and cross power spectra. Does not include CMB spectra.
        Note that P is the deflection angle, but lensing windows return the kappa power.

        :param params: optional :class:`~.model.CAMBparams` instance with parameters to use. If None, must have
          previously set parameters and called :meth:`calc_power_spectra`
          (e.g. if you got this instance using :func:`.camb.get_results`),
        :param lmax: maximum :math:`\ell`
        :param raw_cl: return :math:`C_\ell` rather than :math:`\ell(\ell+1)C_\ell/2\pi`
        :return: dictionary of power spectrum arrays, index as PXP, PxW1, W1xW2, ... etc.
        """

        try:
            if params is not None:
                old_val = params.Want_cl_2D_array
                params.Want_cl_2D_array = True
                self.calc_power_spectra(params)
            elif not self.Params.Want_cl_2D_array:
                raise CAMBValueError('Want_cl_2D_array must be true to have array C_L')
            nwindows = len(self.Params.SourceWindows)
            lmax = lmax or self.Params.max_l
            arr = self.get_unlensed_scalar_array_cls(lmax)
            names = ['P'] + ["W%s" % (i + 1) for i in range(nwindows)]

            result = {}
            for i, name in enumerate(names):
                for j, name2 in enumerate(names):
                    tag = name + 'x' + name2
                    if j < i:
                        result[tag] = result[name2 + 'x' + name]
                    else:
                        cls = arr[i + 2, j + 2, :]
                        if raw_cl:
                            ls = np.arange(1, cls.shape[0])
                            fac = np.float64(ls * (ls + 1))
                            if i == 0 and j == 0:
                                fac *= fac
                            elif i == 0 or j == 0:
                                fac *= np.sqrt(fac)
                            cls[1:] /= (fac / (2 * np.pi))
                        result[tag] = cls
        finally:
            if params is not None:
                params.Want_cl_2D_array = old_val
        return result

    def get_lensed_gradient_cls(self, lmax=None, CMB_unit=None, raw_cl=False, clpp=None):
        r"""
        Get lensed gradient scalar CMB power spectra in flat sky approximation
        (`arXiv:1101.2234 <https://arxiv.org/abs/1101.2234>`_).
        Note that lmax used to calculate results may need to be substantially larger than the lmax
        output from this function (there is no extrapolation as in the main lensing routines).
        Lensed power spectra must be already calculated.

        :param lmax: lmax to output to
        :param CMB_unit: scale results from dimensionless. Use 'muK' for :math:`\mu K^2` units for CMB :math:`C_\ell`
        :param raw_cl: return :math:`C_\ell` rather than :math:`\ell(\ell+1)C_\ell/2\pi`
        :param clpp: custom array of :math:`[L(L+1)]^2 C_L^{\phi\phi}/2\pi` lensing potential power spectrum
             to use (zero based), rather than calculated spectrum from this model
        :return: numpy array CL[0:lmax+1,0:8], where CL[:,i] are :math:`T\nabla T`, :math:`E\nabla E`,
                 :math:`B\nabla B`, :math:`PP_\perp`, :math:`T\nabla E`, :math:`TP_\perp`, :math:`(\nabla T)^2`,
                 :math:`\nabla T\nabla T` where the first six are as defined in appendix C
                 of `1101.2234 <https://arxiv.org/abs/1101.2234>`_.
        """
        assert self.Params.DoLensing
        lmax = self._lmax_setting(lmax)
        res = np.empty((lmax + 1, 8))
        opt = c_int(lmax)
        if clpp is not None:
            if clpp.shape[0] < self.Params.max_l + 1:
                raise CAMBValueError('clpp must go to at least Params.max_l (zero based)')
            clpp = np.array(clpp, dtype=np.float64)
            GetFlatSkyCgrads = lib_import('lensing', '', 'getflatskycgradswithspectrum')
            GetFlatSkyCgrads.argtypes = [POINTER(CAMBdata), numpy_1d, int_arg, numpy_1d]
            GetFlatSkyCgrads(byref(self), clpp, byref(opt), res)
        else:
            GetFlatSkyCgrads = lib_import('lensing', '', 'getflatskycgrads')
            GetFlatSkyCgrads.argtypes = [POINTER(CAMBdata), int_arg, numpy_1d]
            GetFlatSkyCgrads(byref(self), byref(opt), res)
        self._scale_cls(res, CMB_unit, raw_cl)
        return res

    def get_lensed_cls_with_spectrum(self, clpp, lmax=None, CMB_unit=None, raw_cl=False):
        r"""
        Get lensed CMB power spectra using curved-sky correlation function method, using
        cpp as the lensing spectrum. Useful for e.g. getting partially-delensed spectra.

        :param clpp: array of :math:`[L(L+1)]^2 C_L^{\phi\phi}/2\pi` lensing potential power spectrum (zero based)
        :param lmax: lmax to output to
        :param CMB_unit: scale results from dimensionless. Use 'muK' for :math:`\mu K^2` units for CMB :math:`C_\ell`
        :param raw_cl: return :math:`C_\ell` rather than :math:`\ell(\ell+1)C_\ell/2\pi`
        :return: numpy array CL[0:lmax+1,0:4], where 0..3 indexes TT, EE, BB, TE.
        """
        assert self.Params.DoLensing
        lmax_unlens = self.Params.max_l
        if clpp.shape[0] < lmax_unlens + 1:
            raise CAMBValueError('clpp must go to at least Params.max_l (zero based)')
        res = np.zeros((lmax_unlens + 1, 4), dtype=np.float64)
        lmax_lensed = c_int(0)
        lensClsWithSpectrum = lib_import('lensing', '', 'lensclswithspectrum')
        lensClsWithSpectrum.argtypes = [POINTER(CAMBdata), numpy_1d, numpy_2d, int_arg]
        clpp = np.array(clpp, dtype=np.float64)
        lensClsWithSpectrum(byref(self), clpp, res, byref(lmax_lensed))
        res = res[:min(lmax_lensed.value, lmax or lmax_lensed.value) + 1, :]
        self._scale_cls(res, CMB_unit, raw_cl)
        return res

    def get_partially_lensed_cls(self, Alens, lmax=None, CMB_unit=None, raw_cl=False):
        r"""
           Get lensed CMB power spectra using curved-sky correlation function method, using
           true lensing spectrum scaled by Alens. Alens can be an array in L for realistic delensing estimates.
           Note that if Params.Alens is also set, the result is scaled by the product of both

           :param Alens: scaling of the lensing relative to true, with Alens=1 being the standard result.
              Can be a scalar in which case all L are scaled, or a zero-based array with the L by L scaling
              (with L larger than the size of the array having Alens_L=1).
           :param lmax: lmax to output to
           :param CMB_unit: scale results from dimensionless. Use 'muK' for :math:`\mu K^2` units for CMB :math:`C_\ell`
           :param raw_cl: return :math:`C_\ell` rather than :math:`\ell(\ell+1)C_\ell/2\pi`
           :return: numpy array CL[0:lmax+1,0:4], where 0..3 indexes TT, EE, BB, TE.
           """
        clpp = self.get_lens_potential_cls()[:, 0]
        if np.isscalar(Alens):
            clpp *= Alens
        else:
            clpp[:Alens.size] *= Alens
        return self.get_lensed_cls_with_spectrum(clpp, lmax, CMB_unit, raw_cl)

    def angular_diameter_distance(self, z):
        """
        Get (non-comoving) angular diameter distance to redshift z.

        Must have called :meth:`calc_background`, :meth:`calc_background_no_thermo` or calculated transfer
        functions or power spectra.

        :param z: redshift or array of redshifts
        :return: angular diameter distances, matching rank of z
        """
        if np.isscalar(z):
            return self.f_AngularDiameterDistance(byref(c_double(z)))
        else:
            z = np.asarray(z)
            arr = np.empty(z.shape)
            indices = np.argsort(z)
            redshifts = np.array(z[indices], dtype=np.float64)
            self.f_AngularDiameterDistanceArr(arr, redshifts, byref(c_int(z.shape[0])))
            arr[indices] = arr.copy()
            return arr

    def _make_scalar_or_arrays(self, z1, z2):
        if np.isscalar(z1):
            if np.isscalar(z2):
                return z1, z2
            else:
                z1 = np.ones(len(z2)) * z1
        else:
            z1 = np.ascontiguousarray(z1, dtype=np.float64)
            if np.isscalar(z2):
                z2 = np.ones(len(z1)) * z2
            else:
                z2 = np.ascontiguousarray(z2, dtype=np.float64)
                if len(z1) != len(z2):
                    raise CAMBError('z1 nand z2 must be scalar or same-length 1D arrays')
        return z1, z2

    def angular_diameter_distance2(self, z1, z2):
        r"""
        Get angular diameter distance between two redshifts
        :math:`\frac{r}{1+z_2}\text{sin}_K\left(\frac{\chi(z_2) - \chi(z_1)}{r}\right)`
        where :math:`r` is curvature radius and :math:`\chi` is the comoving radial distance.
        If z_1 >= z_2 returns zero.

        Must have called :meth:`calc_background`, :meth:`calc_background_no_thermo` or calculated transfer
        functions or power spectra.

        :param z1: redshift 1, or array of redshifts
        :param z2: redshift 2, or array of redshifts
        :return: result (scalar or array of distances between pairs of z1, z2)
        """
        z1, z2 = self._make_scalar_or_arrays(z1, z2)
        if not np.isscalar(z1):
            dists = np.empty(z1.shape)
            self.f_AngularDiameterDistance2Arr(dists, z1, z2, byref(c_int(dists.shape[0])))
            return dists
        else:
            return self.f_AngularDiameterDistance2(byref(c_double(z1)), byref(c_double(z2)))

    def comoving_radial_distance(self, z, tol=1e-4):
        """
        Get comoving radial distance from us to redshift z in Mpc. This is efficient for arrays.

        Must have called :meth:`calc_background`, :meth:`calc_background_no_thermo` or calculated transfer
        functions or power spectra.

        :param z: redshift
        :param tol: numerical tolerance parameter
        :return: comoving radial distance (Mpc)
        """
        if not np.isscalar(z):
            indices = np.argsort(z)
            redshifts = np.array(z[indices], dtype=np.float64)
            chis = np.empty(redshifts.shape)
            self.f_ComovingRadialDistanceArr(chis, redshifts, byref(c_int(chis.shape[0])), byref(c_double(tol)))
            chis[indices] = chis.copy()
            return chis
        else:
            return self.f_ComovingRadialDistance(byref(c_double(z)))

    def redshift_at_comoving_radial_distance(self, chi):
        """
        Convert comoving radial distance array to redshift array.

        :param chi: comoving radial distance (in Mpc), scalar or array
        :return: redshift at chi, scalar or array
        """
        return self.redshift_at_conformal_time(self.tau0 - chi)

    def redshift_at_conformal_time(self, eta):
        """
        Convert conformal time array to redshift array.
        Note that this function requires the transfers or background to have been calculated with no_thermo=False
        (the default).

        :param eta: conformal time from bing bang (in Mpc), scalar or array
        :return: redshift at eta, scalar or array
        """

        if np.isscalar(eta):
            times = np.array([eta], dtype=np.float64)
        else:
            times = np.ascontiguousarray(eta, dtype=np.float64)
        redshifts = np.empty(times.shape)
        self.f_RedshiftAtTimeArr(redshifts, times, byref(c_int(times.shape[0])))
        config.check_global_error('redshift_at_conformal_time')
        if np.isscalar(eta):
            return redshifts[0]
        else:
            return redshifts

    def luminosity_distance(self, z):
        """
        Get luminosity distance from to redshift z.

        Must have called :meth:`calc_background`, :meth:`calc_background_no_thermo` or calculated transfer functions or
        power spectra.

        :param z: redshift or array of redshifts
        :return: luminosity distance (matches rank of z)
        """

        if not np.isscalar(z):
            z = np.asarray(z)
        return self.angular_diameter_distance(z) * (1.0 + z) ** 2

    def h_of_z(self, z):
        r"""
        Get Hubble rate at redshift z, in :math:`{\rm Mpc}^{-1}` units, scalar or array

        Must have called :meth:`calc_background`, :meth:`calc_background_no_thermo` or calculated transfer functions
        or power spectra.

        Use hubble_parameter instead if you want in [km/s/Mpc] units.

        :param z: redshift
        :return: H(z)
        """
        if not np.isscalar(z):
            z = np.array(z, dtype=np.float64)
            arr = np.empty(z.shape)
            self.f_HofzArr(arr, z, byref(c_int(z.shape[0])))
            return arr
        else:
            return self.f_Hofz(byref(c_double(z)))

    def hubble_parameter(self, z):
        """
        Get Hubble rate at redshift z, in km/s/Mpc units. Scalar or array.

        Must have called :meth:`calc_background`, :meth:`calc_background_no_thermo` or calculated transfer functions
        or power spectra.

        :param z: redshift
        :return: H(z)/[km/s/Mpc]
        """
        return (constants.c / 1e3) * self.h_of_z(z)

    def physical_time_a1_a2(self, a1, a2):
        """
        Get physical time between two scalar factors in Julian Gigayears

        Must have called :meth:`calc_background`, :meth:`calc_background_no_thermo` or calculated transfer functions
        or power spectra.

        :param a1: scale factor 1
        :param a2: scale factor 2
        :return: (age(a2)-age(a1))/Gigayear
        """
        a1, a2 = self._make_scalar_or_arrays(a1, a2)
        if not np.isscalar(a1):
            times = np.empty(a1.shape)
            self.f_DeltaPhysicalTimeGyrArr(times, a1, a2, byref(c_int(times.shape[0])), None)
            return times
        else:
            return self.f_DeltaPhysicalTimeGyr(byref(c_double(a1)), byref(c_double(a2)), None)

    def physical_time(self, z):
        """
        Get physical time from hot big bang to redshift z in Julian Gigayears.

        :param z:  redshift
        :return: t(z)/Gigayear
        """
        if not np.isscalar(z):
            z = np.asarray(z, dtype=np.float64)
        return self.physical_time_a1_a2(0, 1.0 / (1 + z))

    def conformal_time_a1_a2(self, a1, a2):
        """
        Get conformal time between two scale factors (=comoving radial distance travelled by light on light cone)

        :param a1: scale factor 1
        :param a2: scale factor 2
        :return: eta(a2)-eta(a1) = chi(a1)-chi(a2) in Megaparsec
        """
        a1, a2 = self._make_scalar_or_arrays(a1, a2)
        if not np.isscalar(a1):
            times = np.empty(a1.shape)
            self.f_DeltaTimeArr(times, a1, a2, byref(c_int(times.shape[0])), None)
            return times
        else:
            return self.f_DeltaTime(byref(c_double(a1)), byref(c_double(a2)), None)

    def conformal_time(self, z, presorted=None, tol=None):
        """
        Conformal time from hot big bang to redshift z in Megaparsec.

        :param z: redshift or array of redshifts
        :param presorted: if True, redshifts already sorted to be monotonically increasing, if False decreasing,
           or if None unsorted.  If presorted is True or False no checks are done.
        :param tol: integration tolerance
        :return: eta(z)/Mpc
        """
        if np.isscalar(z):
            redshifts = np.array([z], dtype=np.float64)
        else:
            redshifts = np.array(z, dtype=np.float64)
            if presorted is True:
                redshifts = redshifts[::-1].copy()
            elif presorted is None:
                indices = np.argsort(redshifts)[::-1]
                redshifts = redshifts[indices]

        eta = np.empty(redshifts.shape)
        if tol:
            tol = byref(c_double(tol))

        self.f_TimeOfzArr(eta, redshifts, byref(c_int(eta.shape[0])), tol)
        if np.isscalar(z):
            return eta[0]
        else:
            if presorted is False:
                return eta
            elif presorted is True:
                return eta[::-1]
            else:
                eta[indices] = eta.copy()
                return eta

    def sound_horizon(self, z):
        """
        Get comoving sound horizon as function of redshift in Megaparsecs, the integral of the sound speed
        up to given redshift.

        :param z: redshift or array of redshifts
        :return: r_s(z)
        """
        if np.isscalar(z):
            redshifts = np.array([z], dtype=np.float64)
        else:
            redshifts = np.array(z, dtype=np.float64)
        rs = np.empty(redshifts.shape)
        self.f_sound_horizon_zArr(rs, redshifts, byref(c_int(redshifts.shape[0])))
        if np.isscalar(z):
            return rs[0]
        else:
            return rs

    def cosmomc_theta(self):
        r"""
        Get :math:`\theta_{\rm MC}`, an approximation of the ratio of the sound horizon to the angular diameter
        distance at recombination.

        :return: :math:`\theta_{\rm MC}`
        """
        return self.f_CosmomcTheta()


CAMBdata_gettransfers = camblib.__handles_MOD_cambdata_gettransfers
CAMBdata_gettransfers.argtypes = [POINTER(CAMBdata), POINTER(model.CAMBparams),
                                  POINTER(c_int), POINTER(c_int)]
CAMBdata_gettransfers.restype = c_int

CAMBdata_transferstopowers = camblib.__camb_MOD_camb_transferstopowers
CAMBdata_transferstopowers.argtypes = [POINTER(CAMBdata)]

CAMBdata_mattertransferdata = camblib.__handles_MOD_cambdata_mattertransferdata
CAMBdata_mattertransferdata.argtypes = [POINTER(CAMBdata), POINTER(_MatterTransferData)]

CAMBdata_mattertransferks = camblib.__handles_MOD_cambdata_getmattertransferks
CAMBdata_mattertransferks.argtypes = [POINTER(CAMBdata), POINTER(c_int), numpy_1d]

CAMBdata_cltransferdata = camblib.__handles_MOD_cambdata_cltransferdata
CAMBdata_cltransferdata.argtypes = [POINTER(CAMBdata), POINTER(_ClTransferData), int_arg]
CAMBdata_GetLinearMatterPower = camblib.__handles_MOD_cambdata_getlinearmatterpower
CAMBdata_GetLinearMatterPower.argtypes = [POINTER(CAMBdata), numpy_2d, int_arg, int_arg, int_arg]

CAMBdata_GetNonLinearMatterPower = camblib.__handles_MOD_cambdata_getnonlinearmatterpower
CAMBdata_GetNonLinearMatterPower.argtypes = [POINTER(CAMBdata), numpy_2d, int_arg, int_arg, int_arg]

CAMBdata_GetMatterPower = camblib.__handles_MOD_cambdata_getmatterpower
CAMBdata_GetMatterPower.argtypes = [POINTER(CAMBdata), numpy_2d,
                                    d_arg, d_arg, int_arg, int_arg, int_arg]

CAMBdata_GetSigma8 = camblib.__handles_MOD_cambdata_getsigma8
CAMBdata_GetSigma8.argtypes = [POINTER(CAMBdata), numpy_1d, int_arg]

CAMBdata_GetSigmaRArray = camblib.__handles_MOD_cambdata_getsigmararray
CAMBdata_GetSigmaRArray.argtypes = [POINTER(CAMBdata), numpy_2d, numpy_1d, int_arg, numpy_1d_int, int_arg, int_arg,
                                    int_arg]

CAMBdata_CalcBackgroundTheory = camblib.__handles_MOD_cambdata_calcbackgroundtheory
CAMBdata_CalcBackgroundTheory.argtypes = [POINTER(CAMBdata), POINTER(model.CAMBparams)]
CAMBdata_CalcBackgroundTheory.restype = c_int

CAMB_SetTotCls = camblib.__handles_MOD_camb_settotcls
CAMB_SetUnlensedCls = camblib.__handles_MOD_camb_setunlensedcls
CAMB_SetLensPotentialCls = camblib.__handles_MOD_camb_setlenspotentialcls
CAMB_SetUnlensedScalCls = camblib.__handles_MOD_camb_setunlensedscalcls
CAMB_SetLensedScalCls = camblib.__handles_MOD_camb_setlensedscalcls
CAMB_SetTensorCls = camblib.__handles_MOD_camb_settensorcls

_set_cl_args = [POINTER(CAMBdata), int_arg, numpy_1d]

CAMB_SetTotCls.argtypes = _set_cl_args
CAMB_SetUnlensedCls.argtypes = _set_cl_args
CAMB_SetLensPotentialCls.argtypes = _set_cl_args
CAMB_SetUnlensedScalCls.argtypes = _set_cl_args
CAMB_SetTensorCls.argtypes = _set_cl_args
CAMB_SetLensedScalCls.argtypes = _set_cl_args

CAMB_SetUnlensedScalarArray = camblib.__handles_MOD_camb_setunlensedscalararray
CAMB_SetUnlensedScalarArray.argtypes = [POINTER(CAMBdata), int_arg, ndpointer(c_double, flags='F_CONTIGUOUS', ndim=3),
                                        int_arg]

del _set_cl_args

CAMB_TimeEvolution = camblib.__handles_MOD_camb_timeevolution
CAMB_TimeEvolution.restype = c_bool
CAMB_TimeEvolution.argtypes = [POINTER(CAMBdata), int_arg, numpy_1d, int_arg, numpy_1d,
                               int_arg, ndpointer(c_double, flags='C_CONTIGUOUS', ndim=3),
                               int_arg, POINTER(ctypes.c_void_p)]

CAMB_BackgroundThermalEvolution = camblib.__handles_MOD_getbackgroundthermalevolution
CAMB_BackgroundThermalEvolution.argtypes = [POINTER(CAMBdata), int_arg, numpy_1d, numpy_2d]

CAMB_GetBackgroundOutputs = camblib.__handles_MOD_camb_getbackgroundoutputs
CAMB_GetBackgroundOutputs.argtypes = [POINTER(CAMBdata), numpy_1d, int_arg]



# Classes to produce BBN predictions for helium mass fraction and D/H given other parameters
# Fitting formula for Parthenelope via Julien Lesgourgues Dec 2014
# General interpolation tables, with defaults from Parthenope May 2017 (thanks Ofelia Pesanti)
# Use PRIMAT_Yp_DH_Error.dat table for latest from the PRIMAT code (arXiv: 1801.08023, thanks Cyril Pitrou)
# from 1.4.0, delta_nnu values are relative to 3.044, not 3.046

import numpy as np
import os
from scipy.interpolate import RectBivariateSpline
from .constants import m_H, m_He4, default_nnu

# previously (< 1.4.0) default_interpolation_table = 'PArthENoPE_880.2_standard.dat'
default_interpolation_table = 'PRIMAT_Yp_DH_ErrorMC_2021.dat'


def yhe_to_ypBBN(Yp):
    return -4 * m_H * Yp / (Yp * m_He4 - 4 * Yp * m_H - m_He4)


def ypBBN_to_yhe(YBBN):
    return -YBBN * m_He4 / (-YBBN * m_He4 + 4 * YBBN * m_H - 4 * m_H)


class BBNIterpolator(RectBivariateSpline):
    grid: np.ndarray


class BBNPredictor:
    """
    The base class for making BBN predictions for Helium abundance
    """

    def Y_p(self, ombh2, delta_neff=0.):
        r"""
        Get BBN helium nucleon fraction. Must be implemented by extensions.

        :param ombh2: :math:`\Omega_b h^2`
        :param delta_neff:  additional N_eff relative to standard value (of 3.044)
        :return:  Y_p helium nucleon fraction predicted by BBN
        """
        raise Exception('Not implemented')

    def Y_He(self, ombh2, delta_neff=0.):
        r"""
        Get BBN helium mass fraction for CMB code.

        :param ombh2: :math:`\Omega_b h^2`
        :param delta_neff:  additional N_eff relative to standard value (of 3.044)
        :return: Y_He helium mass fraction predicted by BBN
        """
        return ypBBN_to_yhe(self.Y_p(ombh2, delta_neff))


class BBN_table_interpolator(BBNPredictor):
    """
    BBN predictor based on interpolation from a numerical table calculated by a BBN code.

    Tables are supplied for `Parthenope <http://parthenope.na.infn.it/>`_ 2017 (PArthENoPE_880.2_standard.dat),
    similar but with Marucci rates (PArthENoPE_880.2_marcucci.dat),
    `PRIMAT <http://www2.iap.fr/users/pitrou/primat.htm>`_ (PRIMAT_Yp_DH_Error.dat, PRIMAT_Yp_DH_ErrorMC_2021.dat).

    :param interpolation_table: filename of interpolation table to use.
    :param function_of: two variables that determine the interpolation grid (x,y) in the table,
         matching top column label comment. By default ombh2, DeltaN, and function argument names reflect that,
         but can also be used more generally.

    """

    def __init__(self, interpolation_table=default_interpolation_table, function_of=('ombh2', 'DeltaN')):

        if os.sep not in interpolation_table and '/' not in interpolation_table:
            interpolation_table = os.path.normpath(os.path.join(os.path.dirname(__file__), interpolation_table))
        self.interpolation_table = interpolation_table

        comment = None
        with open(interpolation_table) as f:
            for line in f:
                line = line.strip()
                if line:
                    if line[0] == '#':
                        comment = line[1:]
                    else:
                        break
        assert comment
        columns = comment.split()
        ombh2_i = columns.index(function_of[0])
        DeltaN_i = columns.index(function_of[1])

        table = np.loadtxt(interpolation_table)
        deltans = list(np.unique(table[:, DeltaN_i]))
        ombh2s = list(np.unique(table[:, ombh2_i]))
        assert (table.shape[0] == len(ombh2s) * len(deltans))
        self.interpolators = {}
        for i, col in enumerate(columns):
            if i != ombh2_i and i != DeltaN_i and np.count_nonzero(table[:, i]):
                grid = np.zeros((len(ombh2s), len(deltans)))
                for ix in range(table.shape[0]):
                    grid[ombh2s.index(table[ix, ombh2_i]), deltans.index(table[ix, DeltaN_i])] = table[ix, i]
                self.interpolators[col] = BBNIterpolator(ombh2s, deltans, grid)
                self.interpolators[col].grid = grid

        self.ombh2s = ombh2s
        self.deltans = deltans

    def Y_p(self, ombh2, delta_neff=0., grid=False):
        r"""
        Get BBN helium nucleon fraction by intepolation in table.

        :param ombh2: :math:`\Omega_b h^2` (or, more generally, value of function_of[0])
        :param delta_neff:  additional N_eff relative to standard value (of 3.044) (or value of function_of[1])
        :param grid: parameter for :class:`~scipy:scipy.interpolate.RectBivariateSpline` (whether to evaluate the
           results on a grid spanned by the input arrays, or at points specified by the input arrays)
        :return:  Y_p helium nucleon fraction predicted by BBN. Call Y_He() to get mass fraction instead.
        """
        return self.get('Yp^BBN', ombh2, delta_neff, grid)

    def DH(self, ombh2, delta_neff=0., grid=False):
        r"""
        Get deuterium ratio D/H by interpolation in table

        :param ombh2: :math:`\Omega_b h^2` (or, more generally, value of function_of[0])
        :param delta_neff:  additional N_eff relative to standard value (of 3.044) (or value of function_of[1])
        :param grid: parameter for :class:`~scipy:scipy.interpolate.RectBivariateSpline` (whether to evaluate the
           results on a grid spanned by the input arrays, or at points specified by the input arrays)
        :return: D/H
        """
        return self.get('D/H', ombh2, delta_neff, grid)

    def get(self, name, ombh2, delta_neff=0., grid=False):
        r"""
        Get value for variable "name" by interpolation from table (where name is given in the column header comment)
        For example get('sig(D/H)',0.0222,0) to get the error on D/H

        :param name: string name of the parameter, as given in header of interpolation table
        :param ombh2: :math:`\Omega_b h^2` (or, more generally, value of function_of[0])
        :param delta_neff:  additional N_eff relative to standard value (of 3.044) (or value of function_of[1])
        :param grid: parameter for :class:`~scipy:scipy.interpolate.RectBivariateSpline` (whether to evaluate the
           results on a grid spanned by the input arrays, or at points specified by the input arrays)
        :return:  Interpolated value (or grid)
        """
        if name not in self.interpolators:
            raise ValueError('Unknown BBN table column index "%s"' % name)
        res = self.interpolators[name](ombh2, delta_neff, grid=grid)
        if np.isscalar(ombh2) and np.isscalar(delta_neff):
            return np.float64(res)
        return res


class BBN_fitting_parthenope(BBNPredictor):
    """
    Old BBN predictions for Helium abundance using fitting formulae based on Parthenope (pre 2015).
    """

    def __init__(self, tau_neutron=None):
        """
        :param tau_neutron: fitting formula can use different neutron lifetime, defaults to 880.3s if not specified.
        """
        if tau_neutron is None:
            self.taun = 880.3  # Planck 2015 default
        else:
            self.taun = tau_neutron

    def Y_p(self, ombh2, delta_neff=0., tau_neutron=None):
        r"""
        Get BBN helium nucleon fraction.
        # Parthenope fits, as in Planck 2015 papers

        :param ombh2: :math:`\Omega_b h^2`
        :param delta_neff:  additional N_eff relative to standard value (of 3.046 for consistency with Planck)
        :param tau_neutron: neutron lifetime
        :return:  :math:`Y_p^{\rm BBN}` helium nucleon fraction predicted by BBN
        """
        delta_neff = default_nnu + delta_neff - 3.046
        return (
                0.2311 + 0.9502 * ombh2 - 11.27 * ombh2 * ombh2
                + delta_neff * (0.01356 + 0.008581 * ombh2 - 0.1810 * ombh2 * ombh2)
                + delta_neff * delta_neff * (-0.0009795 - 0.001370 * ombh2 + 0.01746 * ombh2 * ombh2)
        ) * pow((tau_neutron or self.taun) / 880.3, 0.728)

    def DH(self, ombh2, delta_neff, tau_neutron=None):
        delta_neff = default_nnu + delta_neff - 3.046
        return (
                18.754 - 1534.4 * ombh2 + 48656. * ombh2 * ombh2 - 552670. * ombh2 ** 3
                + delta_neff * (2.4914 - 208.11 * ombh2 + 6760.9 * ombh2 ** 2 - 78007. * ombh2 ** 3)
                + delta_neff * delta_neff * (
                        0.012907 - 1.3653 * ombh2 + 37.388 * ombh2 ** 2 - 267.78 * ombh2 ** 3)
        ) * pow((tau_neutron or self.taun) / 880.3, 0.418) * 1e-5


_predictors = {}


def get_predictor(predictor_name=None):
    """
    Get instance of default BBNPredictor class. Currently numerical table interpolation as Planck 2018 analysis.
    """
    global _predictors
    predictor_name = predictor_name or default_interpolation_table
    predictor = _predictors.get(predictor_name, None)
    if predictor is None:
        if predictor_name == 'BBN_fitting_parthenope':
            predictor = BBN_fitting_parthenope()
        else:
            predictor = BBN_table_interpolator(interpolation_table=predictor_name)
        _predictors[predictor_name] = predictor
    return predictor


if __name__ == "__main__":
    print(BBN_table_interpolator().Y_He(0.01897, 1.2))
    print(BBN_fitting_parthenope().Y_He(0.01897, 1.2))
    print(BBN_table_interpolator('PRIMAT_Yp_DH_Error.dat').Y_He(0.01897, 1.2))
    print(BBN_table_interpolator().DH(0.02463, -0.6))
    print(BBN_fitting_parthenope().DH(0.02463, -0.6))
    print(BBN_table_interpolator('PArthENoPE_880.2_marcucci.dat').DH(0.02463, -0.6))
    print(BBN_table_interpolator('PRIMAT_Yp_DH_Error.dat').DH(0.02463, -0.6))
    print(BBN_table_interpolator('PRIMAT_Yp_DH_ErrorMC_2021.dat').DH(0.02463, -0.6))
